/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file failstop.h
 *
 * @brief Macros, Data Types, Global/Static Variables, and Function
 * Prototypes/Implementations for a fail-stop failure model.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#ifndef __FAILSTOP_H_
#define __FAILSTOP_H_

#include <assert.h>

#include "log.h"



/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

/**
 * @brief Do Or Die
 *
 * If expr is false, then this macro terminates the process.
 *
 * This is intended for use with any functions that are not expected to fail in
 * the normal course of execution.
 */
#define dod(expr)                                       \
    do {                                                \
        if (!(expr)) {                                  \
            log(LOG_EMU_ERRORS | LOG_SFS_ERRORS,        \
                "Error on line %d in %s: %s\n",         \
                __LINE__, __FILE__, strerror(errno));   \
            assert(0);                                  \
        }                                               \
    } while(0)


#define CHECK_ERROR(err_func)                                           \
    do {                                                                \
        if (retval < 0) {                                               \
            retval = -errno;                                            \
            log(LOG_SFS_ERRORS, "Error : " err_func " %s "              \
                "at %s:%d\n", strerror(errno), __FILE__, __LINE__);     \
            return retval;                                              \
        }                                                               \
    } while(0)

#define CHECK_ERROR_NULL(err_func)                                      \
    do {                                                                \
        if (retval == 0) {                                              \
            retval = retval - errno;                                    \
            log(LOG_SFS_ERRORS, "Error : " err_func " %s "              \
                "at %s:%d\n", strerror(errno), __FILE__, __LINE__);     \
            return retval;                                              \
        }                                                               \
    } while(0)

#endif  /* __FAILSTOP_H_ */
