/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file log.h
 *
 * @brief Macros, data structures, and function prototypes for implementing
 * logging.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 *
 * @bugs No known bugs
 */

#ifndef __LOG_H_
#define __LOG_H_

#include <sys/stat.h>



/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */


#define LOG_NONE                0x00000000
#define LOG_ALL                 0xFFFFFFFF

#define LOG_EMU_ALL             0x000000FF
#define LOG_EMU_CALLS   0x00000001
#define LOG_EMU_ERRORS  0x00000002
#define LOG_EMU_INFO    0x00000004
#define LOG_EMU_STATS   0x00000008
#define LOG_EMU_RETURN  0x00000010

#define LOG_SFS_ALL             0x0000FF00
#define LOG_SFS_CALLS   0x00000100
#define LOG_SFS_ERRORS  0x00000200
#define LOG_SFS_INFO    0x00000400
#define LOG_SFS_STATS   0x00000800
#define LOG_SFS_RETURN  0x00001000

#define LOG_BC_ALL              0x00FF0000
#define LOG_BC_CALLS    0x00010000
#define LOG_BC_ERRORS   0x00020000
#define LOG_BC_INFO             0x00040000
#define LOG_BC_STATS    0x00080000
#define LOG_BC_RETURN   0x00100000

#define LOG_WRK_ALL             0x0F000000
#define LOG_WRK_ERRORS  0x02000000
#define LOG_WRK_INFO    0x04000000
#define LOG_WRK_STATS   0x08000000

#define LOG_CLN_ALL             0xF0000000
#define LOG_CLN_ERRORS  0x02000000
#define LOG_CLN_INFO    0x04000000
#define LOG_CLN_STATS   0x80000000


#define log_verbosity   LOG_ALL & (~LOG_EMU_ALL)


#ifdef GENERATE_LOGS
#define log_open(arg)                   smr_log_open(arg)
#define log_close()                             smr_log_close
#define log_stats(path, buf)    smr_log_stat(__FUNCTION__, path, buf)
#define log(level, ...)                                 \
    do {                                                \
        if ((level) & (log_verbosity))                  \
            smr_log_msg(__FUNCTION__, __VA_ARGS__);     \
    } while(0)
#else
#define log_open(arg)
#define log_close()
#define log_stats(path, buf)
#define log(level, ...)
#endif



/* ====================== */
/* -- Static Variables -- */
/* ====================== */





/* ========================= */
/* -- Function Prototypes -- */
/* ========================= */

    void smr_log_open(const char *log_path);
void smr_log_msg(const char *function, const char *format, ...);
void smr_log_stat(const char *function, const char *path,
                  const struct stat *statbuf);
void smr_log_close();



#endif      /* __LOG_H_ */
