/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file band_bitmap.c
 *
 * @brief Implementation of functionality pertaining to Bands
 *
 * @author Jainam Shah <jashah@andrew.cmu.edu>
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#define _XOPEN_SOURCE 500
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "log.h"
#include "failstop.h"
#include "common.h"
#include "helper.h"
#include "emulator.h"
#include "band_bitmap.h"


#define FILE_BAND_BITMAP                                "/.smr_band_bitmap"

static int fd;

open_band_t open_band;

/*
 * @brief Initializes the band bitmap.
 *
 * If the band bitmap file exists, then it will open the bitmap file.
 *
 * @return  0 on success; -1 on failure
 */

int init_band_bitmap(void)
{
    int i;
    char path[PATH_MAX];
    char bit = 0;
    int band_modulocount;

    /* Assumption : Band count is a multiple of 8.
     * (This is fixed. Need testing though) */

    /* Identify the no. of bands on the emulated disk */
    band_modulocount = (SFSDATA.edi_bandcount + 7) / 8;


    /* Initialize the last open band */
    open_band.band = 0;
    open_band.rba = 0;
    open_band.total_blocks = 0;


    /* Open the Band Bitmap file */
    translate_path(FILE_BAND_BITMAP, SFSDATA.path_unshingled, path);
    fd = open(path, O_RDWR | O_CREAT | O_EXCL, 0644);
    if (fd < 0) {
        // Band bitmap file already exists.
        dod((fd = open(path, O_RDWR)) > 0);
        return 0;
    }


    /* If the file is created then initialize the bitmap */
    for (i = 0; i < band_modulocount; i++) {
        dod(pwrite(fd, &bit, 1, i) == 1);
    }

    return 0;
}


/*
 * @brief Get next free band and update the band bitmap.
 *
 * @return Free band number
 */

int get_freebandnum()
{
    int i, j;
    int max_bands;
    int band_num = 0;
    char bit;


    max_bands = (SFSDATA.edi_bandcount + 7) / 8;

    for (i = 0; i < max_bands; i++) {
        dod(pread(fd, &bit, 1, i) == 1);

        if(bit == 0xFF) {
            continue;
        }

        for (j = 0; j < 8; j++) {
            if ((bit & (1 << j)) == 0) {
                bit = bit | (1 << j);
                dod(pwrite(fd, &bit, 1, i) == 1);
                band_num = i * 8 + j + 1;
                break;
            }
        }
        if (band_num != 0) {
            break;
        }
    }

    /* For the last set of bands, since we rounded to 8,
     * the band number may be over the available count */
    if(band_num > SFSDATA.edi_bandcount) {
        return 0;
    }

    return band_num;
}


/*
 * @brief Set the band by updating the band bitmap.
 *
 * @param band_num: Band number of the band to be set in the bitmap.
 */

void put_setbandnum(int band_num)
{
    int i, j;
    char bit;

    if (band_num <= 0) {
        return;
    }

    i = (band_num - 1) / 8;
    j = (band_num - 1) % 8;

    dod(pread(fd, &bit, 1, i) == 1);

    bit = bit | (1 << j);

    dod(pwrite(fd, &bit, 1, i) == 1);
}


/*
 * @brief Free the band by updating the band bitmap
 *
 * @param band_num: Band number of the band to be freed
 */

void put_freebandnum(int band_num)
{
    int i, j;
    char bit;

    if (band_num <= 0) {
        return;
    }

    i = (band_num - 1) / 8;
    j = (band_num - 1) % 8;

    dod(pread(fd, &bit, 1, i) == 1);

    bit = bit & (~(1 << j));

    dod(pwrite(fd, &bit, 1, i) == 1);
}


/*
 * @brief Test if the band is free or occupied.
 *
 * @param band_num : Band number.
 *
 * @return  0 : if Band is free.
 *          1 : if Band is occupied.
 *         -1 : on Error.
 */

int test_bandusage(int band_num)
{
    int i, j;
    char bit;
    int usage = 0;

    if (band_num <= 0) {
        return -1;
    }

    i = (band_num - 1) / 8;
    j = (band_num - 1) % 8;

    dod(pread(fd, &bit, 1, i) == 1);

    usage = bit & (1 << j);
    if (usage == 0) {
        return 0;
    }

    return 1;
}


/*
 * @brief Retrieves the next band that has at least block_count free blocks.
 *
 * @param edi_fd File descriptor to EDI file
 * @param band Pointer to open_band_t structure
 * @param block_count Min. no. of blocks required in the band
 *
 * @return 0 on success; -1 on failure
 */

int get_openband(int edi_meta_fd, open_band_t *band, int block_count)
{
    /* Read the no. of blocks for a band */
    if (band->total_blocks == 0) {
        band->total_blocks = SFSDATA.edi_bandsize;
    }

    if (band->band != 0) {
        dod(!edi_managebands(edi_meta_fd, SVC_GETPTR, &band->band, &band->rba, 1));

        /* Check if the band has enough blocks */
        if (band->total_blocks - band->rba >= block_count) {
            return 0;
        }
    }

    band->band = get_freebandnum();
    band->rba = 0;

    if (band->band != 0) {
        return 0;
    }

    for (band->band = 1; band->band <= SFSDATA.edi_bandcount; (band->band)++) {
        /* Get the write pointer for the current band */
        dod(!edi_managebands(edi_meta_fd, SVC_GETPTR, &band->band, &band->rba, 1));

        /* Check if the current band has enough space */
        if (band->total_blocks - band->rba >= block_count) {
            return 0;
        }
    }

    return -1;
}
