/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>


#include "helper.h"
#include "common.h"
#include "errno.h"
#include "string.h"

#include <stdlib.h>


#include "band_log.h"
#include "band_meminfo.h"

#define FILE_BANDLOG_APPEND_PREFIX              "/.smr_append_log_"
#define FILE_BANDLOG_DELETE_PREFIX              "/.smr_delete_log_"


/*
 * @brief : Initialize the store log for the given band number.
 *
 * @param band_num : Band number.
 *
 * @return fd : File descriptor.
 *         -ve number : on Error.
 *
 */

int init_storelog(int band_num)
{
    int fd;
    char band_name[PATH_MAX];
    char path[PATH_MAX];

    sprintf(band_name, "%s%d", FILE_BANDLOG_APPEND_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    fd = open(path, O_RDWR | O_APPEND | O_CREAT | O_EXCL, 0644);

    if (fd < 0) {
        //append log file already exist.
        fd = open(path, O_RDWR | O_APPEND);
        if (fd < 0) {
            dbg("Error opening store log: %s \n", strerror(errno));
            return -1;
        }
        return fd;
    }

    return fd;
}


/*
 * @brief : Initialize the delete log for a band.
 *
 * @param band_num : Band number.
 *
 * @return fd : File descriptor.
 *         -ve number : on Error.
 *
 */

int init_deletelog(int band_num)
{
    int fd;
    char band_name[PATH_MAX];
    char path[PATH_MAX];

    sprintf(band_name, "%s%d", FILE_BANDLOG_DELETE_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    fd = open(path, O_RDWR | O_APPEND | O_CREAT | O_EXCL, 0644);

    if (fd < 0) {
        //delete log file already exist.
        fd = open(path, O_RDWR | O_APPEND);  
        if (fd < 0) {
            dbg("Error opening delete log  \n");
            return -1;
        }
        return fd;
    }

    return fd;
}


/*
 * @brief : Deletes the store log for given band.
 *
 * @param band_num : Band number.
 *
 * @return : None.
 *
 */

void clear_storelog(int band_num)
{
    char band_name[PATH_MAX];
    char path[PATH_MAX];

    sprintf(band_name, "%s%d", FILE_BANDLOG_APPEND_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);
 
    unlink(path);
}


/*
 * @brief : Deletes the delete log for given band. Is called when a band is
 *          cleaned.
 *
 * @param band_num : Band number.
 * @param path_unshingled : Path to the unshingled drive.
 *
 * @return : None.
 *
 */

void clear_deletelog(int band_num)
{
    char band_name[PATH_MAX];
    char path[PATH_MAX];

    sprintf(band_name, "%s%d", FILE_BANDLOG_DELETE_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    unlink(path);
}


/*
 * @brief : Appends an entry to the store log.
 *
 * @param fd : File descriptor.
 * @param offset : Offset in the band where the file is stored.
 * @param size : Size of the file.
 * @param inode_num : inode number of the file.
 *
 * @return : Index of the appended entry.
 *
 */

int  append_storelog(band_t band_num, int fd, int offset, 
                     int size, int inode_num, long int age)
{
    struct store_entry entry;
    struct stat filestat;
    int entrynumber;

    entry.offset = offset;
    entry.size = size;
    entry.inode_num = inode_num;
    entry.age = age;

    if (write(fd, &entry, sizeof(struct store_entry)) < 0) {
        dbg("Error writing to the append log in append_storelog \n");
        exit(1);
    }

    band_meminfo_increment_blocks_used(band_num, size);
    
    //Calculate the index of the insertion using the file size.
    fstat(fd, &filestat);
    entrynumber = filestat.st_size / (sizeof(struct store_entry));

    return entrynumber;
}


/*
 * @brief : Appends an entry to the delete log.
 *          Need to grab the lock for the band before increment blocks
 * @param fd : File descriptor.
 * @param offset : Offset in the band where the file was stored.
 * @param size : Size of the file.
 * @param entrynumber : Index in the append log.
 *
 * @return : None.
 */

void append_deletelog(band_t band_num, int fd, int offset, int size, int entrynumber)
{
    struct delete_entry entry;;

    entry.offset = offset;
    entry.size = size;
    entry.entrynumber = entrynumber;

    if (write(fd, &entry, sizeof(struct delete_entry)) < 0) {
        dbg("Error writing delete log in append_deletelog\n");
        exit(1);
    }

    band_meminfo_increment_blocks_deleted(band_num, size);
}


/*
 * @brief : List the entries of the band where the data is live on the band.
 *
 * @param band_num : Band number.
 * @param store_fd : File Descriptor for append log.
 * @param delete_fd : File Descriptor for delete log.
 * @param length : Pointer to store the number of entries returned.
 * @param livedatasize : Number of live block in the band.
 *
 * @return : List of struct containing append log entries not present in delete
 *           log. The list of structure is malloced here and needs tobe freed 
 *           by the user afterwards.
 *
 */

struct store_entry*  get_livelog(int band_num, int store_fd, int delete_fd,
                                 int *length, int *livedatablocks)
{
    struct store_entry *live;
    struct delete_entry del;
    struct store_entry entry;
    int *deleted;
    int numentries;
    int delentries;
    int i;
    int livecount = 0;
    int liveblock = 0;

    struct stat filestat;

    fstat(store_fd, &filestat);
    numentries = filestat.st_size / sizeof(struct store_entry);

    fstat(delete_fd, &filestat);
    delentries = filestat.st_size / sizeof(struct delete_entry);

    deleted = (int *) calloc (numentries, sizeof(int));
    if (deleted == NULL) {
        dbg("Memory allocation failed in get_livelog\n");
        exit(1);
    }

    //Read the delete log and mark the entires which are deleted in a array.
    for (i = 0; i < delentries; i++) {
        pread(delete_fd, &del, sizeof(struct delete_entry), 
              i * sizeof(struct delete_entry));

        if (del.entrynumber > numentries) {
            dbg("wrong value of delete entry \n");
            exit(1);
        }

        deleted[del.entrynumber - 1] = 1;
    }

    live = (struct store_entry *) calloc((numentries - delentries),
                                         sizeof(struct store_entry));
    if (live == NULL) {
        dbg("Malloc failed \n");
        exit(1);
    }

    //Now using the array, read only those entries which are not present in the
    //delete log.
    for (i = 0; i < numentries; i++) {
        if (deleted[i] == 0) {
            if (pread(store_fd, &entry, sizeof(struct store_entry),
                      i * sizeof(struct store_entry)) < 0) {
                dbg("Error reading the store log in get_livelog \n");
            }

            live[livecount] = entry;
            liveblock = liveblock + entry.size;
            livecount++;
        }
    }

    *length = livecount;
    *livedatablocks = liveblock;
    free(deleted);
    return live;

}


/*
 * @brief : Returns the size of the band currently used.
 *
 * @param band_num : Band number.
 *
 * @return size : Size of the band used.
 *            -1 : on Error.
 *
 */

int band_occupancy(int band_num)
{
    int fd;
    char band_name[PATH_MAX];
    char path[PATH_MAX];
    struct store_entry lastentry;
    int error;

    assert(0); /* use meminfo instead */

    sprintf(band_name, "%s%d", FILE_BANDLOG_APPEND_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    fd = open(path, O_RDWR);
    if (fd < 0) {
        dbg("Error opening store log file in band_occupancy \n");
        return -1;
    }

    lseek(fd, -1 * (sizeof(struct store_entry)), SEEK_END);
    error = read(fd, &lastentry, sizeof(struct store_entry));

    close(fd);
    if (error < 0) {
        dbg("Error reading the store log in band_occupancy \n");
        return -1;
    }

    if (error == 0) {
        return 0;
    }

    return lastentry.offset + lastentry.size;
}


/*
 * @brief : Returns the age of the band currently used.
 *
 * @param band_num : Band number.
 *
 * @return age : Age of the band.
 *            -1 : on Error.
 *
 */

int band_age(int band_num)
{
    int fd;
    char band_name[PATH_MAX];
    char path[PATH_MAX];
    struct store_entry firstentry;
    int error;

    //TODO: Currently returns the age of the first file.
    //Needs to parse the entire log to get the age of the band.

    sprintf(band_name, "%s%d", FILE_BANDLOG_APPEND_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    fd = open(path, O_RDWR);
    if (fd < 0) {
        dbg("Error opening store log in band_age \n");
        return -1;
    }

    error = read(fd, &firstentry, sizeof(struct store_entry));

    close(fd);
    if (error < 0) {
        dbg("Error reading append log in band_age \n");
        return -1;
    }

    if (error == 0) {
        return 0;
    }

    return firstentry.age;
}


/*
 * @brief : Return the number of entries in the delete log.
 *          Function currently used in cleaning for the base line
 *          cleaning thread.
 * 
 * @param band_num : Band number.
 *
 * @return num_entries : Number of entries in delete log.
 *
 */

int band_delentries(int band_num)
{
    char band_name[PATH_MAX];
    char path[PATH_MAX];
    struct stat filestat;
    int error;
    int num_entries;

    sprintf(band_name, "%s%d", FILE_BANDLOG_DELETE_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    error = stat(path, &filestat);
    if (error == -1) {
        if (errno == ENOENT) {
            //delete log does not exist for file
            return 0;
        }
        return 0;
    }

    num_entries = filestat.st_size / sizeof (struct delete_entry);
    return num_entries;
}

void get_livelog_wrapper(band_t band_frag, struct store_entry **live,
        int *length) {
    int frag_storefd;
    int frag_deletefd;
    int livedata;
    frag_storefd = init_storelog(band_frag);
    if (frag_storefd == -1) {
        exit(EXIT_FAILURE);
    }

    frag_deletefd = init_deletelog(band_frag);
    if (frag_deletefd == -1) {
        exit(EXIT_FAILURE);
    }

    *live = get_livelog(band_frag, frag_storefd, frag_deletefd, length,
                       &livedata);

    close(frag_storefd);
    close(frag_deletefd);
    return;
}


int recover_storelog(int band_num)
{
    int fd, i;
    char band_name[PATH_MAX];
    char path[PATH_MAX];
    int numentries;
    struct stat filestat;
    struct store_entry entry;
    size_t size = 0;

    sprintf(band_name, "%s%d", FILE_BANDLOG_APPEND_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    fd = open(path, O_RDONLY);
    if(fd < 0) {
        /* If the log file is not present, then we need not recover any data */
        return 0;
    }

    fstat(fd, &filestat);
    numentries = filestat.st_size / sizeof(struct store_entry);

    for (i = 0; i < numentries; i++) {
        if(pread(fd, &entry, sizeof(struct store_entry), 
              i * sizeof(struct store_entry)) < 0) {
            dbg("Error reading the store log for recover\n");
            close(fd);
            return -1;
        }

        band_meminfo_increment_blocks_used(band_num, entry.size);

        size += entry.size;
    }

    log(LOG_BC_INFO, "Read storelog of band %d: %d blocks used\n",
                    band_num, size);
            
    close(fd);
    return 0;
}


int recover_deletelog(int band_num)
{
    int fd, i;
    char band_name[PATH_MAX];
    char path[PATH_MAX];
    int numentries;
    struct stat filestat;
    struct delete_entry entry;
    size_t size = 0;

    sprintf(band_name, "%s%d", FILE_BANDLOG_DELETE_PREFIX, band_num);
    translate_path(band_name, SFSDATA.path_unshingled, path);

    fd = open(path, O_RDONLY);
    if(fd < 0) {
        /* If the log file is not present, then we need not recover any data */
        return 0;
    }

    fstat(fd, &filestat);
    numentries = filestat.st_size / sizeof(struct delete_entry);

    for (i = 0; i < numentries; i++) {
        if(pread(fd, &entry, sizeof(struct delete_entry), 
              i * sizeof(struct delete_entry)) < 0) {
            dbg("Error reading the delete log for recover\n");
            close(fd);
            return -1;
        }

        band_meminfo_increment_blocks_deleted(band_num, entry.size);

        size += entry.size;
    }
     
    log(LOG_BC_INFO, "Read deletelog of band %d: %d blocks deleted\n",
                    band_num, size);
            
    close(fd);
    return 0;
}


int recover_band_log_data()
{
    int band_num;

    for(band_num = 0; band_num < SFSDATA.edi_bandcount; band_num++) {
        if(recover_storelog(band_num) < 0)
            return -1;

        if(recover_deletelog(band_num) < 0)
            return -1;
    }

    return 0;
}
