/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file edi_stats.c
 *
 * @brief Macros, Data Types, Global/Static Variables, and Function 
 * Prototypes/Implementations pertaining to the EDI Statistics Generator
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>

#include "debug.h"
#include "common.h"
#include "helper.h"
#include "emulator.h"
#include "band_bitmap.h"
#include "band_log.h"


/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

#define ARG_COUNT_MIN                   4

#define ARG_PATH_EDI                    1
#define ARG_PATH_USHNG                  2
#define ARG_PATH_OUTPUT                 3
#define ARG_PATH_STATS                  4



/* ====================== */
/* -- Type Definitions -- */
/* ====================== */



/* ============================= */
/* -- Global/Static Variables -- */
/* ============================= */

FILE            *file_stats;
FILE            *path_stats;
static char path_stats_file[PATH_MAX];
static int  per_band_stats;
shingledfs_data_t shingledfs_data;


/* ========================= */
/* -- Function Prototypes -- */
/* ========================= */

static void initialize();
static void process_band(band_t band);



/* ============================== */
/* -- Function Implementations -- */
/* ============================== */

/**
 * @brief Prints the usage format
 */
static void usage(const char *program_name)
{
    fprintf(stderr,
            "USAGE: %s <EDI path> <Unshingled path> <Output path> [Per band stats output path]\n",
            program_name);
    exit(1);
}



/**
 * @brief Processes the incoming arguments.
 *
 * This function stores the required paths, namely the path to the unshingled
 * partition, the EDI file and the Output file for later use.
 *
 * If this function finds an error with the arguments, then it terminates the
 * application.
 *
 * @param argc Pointer to the count of arguments
 * @param argv Array of command line arguments
 */
static void process_args(int *argc, char *argv[]) {
    int i, arg_error = 0;


    /* Ensure that the arguments are available */
    if (*argc < ARG_COUNT_MIN)
        usage(argv[0]);


    /* Strip out any terminating /'s in the path to the directories */
    i = strlen(argv[ARG_PATH_USHNG]) - 1;
    if (argv[ARG_PATH_USHNG][i] == '/')
        argv[ARG_PATH_USHNG][i] = '\0';


    /* Validate the path to the Emulated Disk Image */
    if (!is_valid_file(argv[ARG_PATH_EDI])) {
        arg_error = 1;
        fprintf(stderr, "%s is not a valid file!\n", argv[ARG_PATH_EDI]);
    }

    /* Validate the path to the Unshingled partition */
    if (!is_valid_dir(argv[ARG_PATH_USHNG])) {
        arg_error = 1;
        fprintf(stderr, "%s is not a valid directory!\n", argv[ARG_PATH_USHNG]);
    }

    if (*argc == 5) {
        per_band_stats = 1;
        strncpy(path_stats_file, argv[ARG_PATH_STATS], sizeof path_stats_file);
    }

    /* Check if there are any errors */
    if (arg_error)
        usage(argv[0]);


    /* Store the arguments */
    strcpy(SFSDATA.path_edi, argv[ARG_PATH_EDI]);
    strcpy(SFSDATA.path_unshingled, argv[ARG_PATH_USHNG]);
    strcpy(SFSDATA.path_tmpfs, argv[ARG_PATH_OUTPUT]);

    dbg("EDI path: %s\n", SFSDATA.path_edi);
    dbg("Unshingled path: %s\n", SFSDATA.path_unshingled);
    dbg("Output path: %s\n", SFSDATA.path_tmpfs);
    if (per_band_stats) {
        dbg("Stats file: %s\n", path_stats_file);
    }
}



/**
 * @brief Initializes the application by opening the EDI and the output file
 *
 * This function terminates the program if there are problems opening the EDI
 * file or the output file.
 */
static void initialize() {
    /* Load the Emulated Disk Image */
    if (edi_init()) {
        fprintf(stderr, "Error initializing Emulator interface!\n");
        exit(EXIT_FAILURE);
    }

#ifdef _EDI_ON_RAW
    if ((SFSDATA.fd_edi = edi_open(SFSDATA.path_edi_metadata, SFSDATA.path_edi, &SFSDATA.fd_edi_metadata)) < 0) {
#else
    if ((SFSDATA.fd_edi = edi_open(SFSDATA.path_edi)) < 0) {
#endif
        fprintf(stderr, "edi_open(%s): %s!\n", SFSDATA.path_edi, strerror(errno));
        exit(EXIT_FAILURE);
    }

    /* Open the output file */
    if ((file_stats = fopen(SFSDATA.path_tmpfs, "w+")) == NULL) {
        fprintf(stderr, "fopen(%s): %s\n", SFSDATA.path_tmpfs, strerror(errno));
        exit(EXIT_FAILURE);
    }


    if (per_band_stats) {
        /* Open per band stats file, if user requests it */
        if ((path_stats = fopen(path_stats_file, "a+")) == NULL) {
            fprintf(stderr, "fopen(%s): %s\n", path_stats_file, strerror(errno));
            exit(EXIT_FAILURE);
        }
    }

    /* Read the EDI Parameters */
    read_edi_parameters(SFSDATA.fd_edi, &SFSDATA.edi_blksize,
                        &SFSDATA.edi_bandcount, &SFSDATA.edi_bandsize);

    /* TODO: Init band_meminfo from append and delete log */
}



/**
 * @brief Generates statistics for the specified band
 *
 * This function writes output to the output file directly
 *
 * @param band The band no. of the band to generate statistics for
 */
static void process_band(band_t band) {
    char error[PATH_MAX];
    int i, band_used = 0, fd_append, fd_delete, num_append, num_delete;
    rba_t band_ptr = 0;
    struct stat statbuf;
    int *deleted;
    struct delete_entry rec_delete;
    struct store_entry rec_append;


    /* Read the write pointer for the band */
    if (edi_managebands(SFSDATA.fd_edi, SVC_GETPTR, &band, &band_ptr, 0)) {
        sprintf(error, "edi_managebands() failed!");
        goto error;
    }

    if(band_ptr > 0)
        band_used = 1;
    dbg("Band State: %s\n", band_used ? "Used" : "Free");


    /* Open the Append Log and the Delete Log for the band */
    if ((fd_append = init_storelog(band)) < 0) {
        sprintf(error, "Error opening Append Log!");
        goto error;
    }

    if ((fd_delete = init_deletelog(band)) < 0) {
        sprintf(error, "Error opening Delete Log!");
        goto error;
    }


    int blocks_used = 0;
    int blocks_dead = 0;
    int blocks_free = 0;
    /* Generate statistics for the band if it is used */
    if (band_used) {
        /* Read the no. of append and delete log entries */
        fstat(fd_append, &statbuf);
        num_append = statbuf.st_size / sizeof(struct store_entry);
        dbg("Append Log Entries:%d\n",num_append);

        fstat(fd_delete, &statbuf);
        num_delete = statbuf.st_size / sizeof(struct delete_entry);
        dbg("Delete Log Entries:%d\n",num_delete);


        /* Identify deleted append log entries from the delete log */
        deleted = calloc(num_append, sizeof(int));
        for (i=0; i<num_delete; i++) {
            dbg("Processing Delete Entry %d\n", i);
            pread(fd_delete, &rec_delete, sizeof(struct delete_entry),
                  i * sizeof(struct delete_entry));

            if (rec_delete.entrynumber > num_append) {
                sprintf(error, "Delete Log entry %d refers to %d of %d Append"
                        "Log entries!", i, rec_delete.entrynumber, num_append);
                goto error;
            }

            deleted[rec_delete.entrynumber - 1] = 1;
            //band_dead += rec_delete.size;
        }


        /* Identify the live data in the band */
        for (i=0; i<num_append; i++) {
           dbg("Processing Append Entry %d\n", i);
            pread(fd_append, &rec_append, sizeof(struct store_entry),
                  i * sizeof(struct store_entry));

            if (deleted[i]) {
                blocks_dead += rec_append.size;
            } else {
                //band_live += rec_append.size;
                blocks_used += rec_append.size;
            }
        }
    } 
    /* account for the unused free blocks at the end of band */
    blocks_free = (SFSDATA.edi_bandsize - band_ptr);

    //assert (((blocks_dead + blocks_used) - band_ptr < 10) ||
    //        (band_ptr - (blocks_dead + blocks_used) < 10));
    assert(blocks_dead + blocks_used == band_ptr);
    if (per_band_stats) {
        fprintf(path_stats, "%d\t%d\t%d\t", blocks_used, blocks_dead, 
                                        blocks_free);
    }

    /* Close the open files */
    close(fd_append);
    close(fd_delete);


    /* Write the band statistics to the output file */
    fprintf(file_stats, "Band %3d\t%s\t%8d\t%8d\t%8d\t%8d\t%8d\n",
            band, (band_used ? "Used" : "Free"), SFSDATA.edi_bandsize,
            band_ptr, blocks_used, blocks_free, blocks_dead);

    return;

error:
    fprintf(file_stats, "Band %3d: %s\n", band, error);
}



/* =================== */
/* -- Main Function -- */
/* =================== */

int main(int argc, char *argv[]) {
    int i;
    time_t curtime = time(NULL);


    /* Process the arguments */
    process_args(&argc, argv);

    /* Initialize the program */
    initialize();


    /* Output the basic EDI parameters */
    fprintf(file_stats, "EDI Image: %s\n", SFSDATA.path_edi);
    fprintf(file_stats, "Unshingled Partition: %s\n", SFSDATA.path_unshingled);
    fprintf(file_stats, "Report Time: %s\n", ctime(&curtime));
    fprintf(file_stats, "Sector Size: %d bytes\n", SFSDATA.edi_blksize);
    fprintf(file_stats, "Bands: %d bytes\n", SFSDATA.edi_bandcount);
    fprintf(file_stats, "Band Size: %d blocks (%d bytes)\n",
            SFSDATA.edi_bandsize, SFSDATA.edi_bandsize * SFSDATA.edi_blksize);


    /* Iterate over the bands in the EDI */
    fprintf(file_stats, "Band No.\tState\tBand Size\tCursor Position\t"
            "Used Blocks\tFree Blocks\tDead Blocks\n");

    if (per_band_stats) {
        fprintf(path_stats, "%lu\t", time(NULL));
    }

    for (i=1; i<SFSDATA.edi_bandcount; i++) {
        dbg("Processing band %d\n", i);
        process_band(i);
    }


    /* Close the output file */
    fclose(file_stats);

    if (per_band_stats) {
        fprintf(path_stats, "\n");
        fclose(path_stats);
    }
    return (EXIT_SUCCESS);
}
