/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file fuseops.c
 *
 * @brief Implementation of the FUSE operations needed by Shingled FS.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 *
 * @bug shingledfs_link() is bound to fail under certain conditions. @see
 * shingledfs_link for more details
 */

#define _XOPEN_SOURCE 500
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>
#include <dirent.h>
#include <assert.h>
#include <sys/xattr.h>
#include <sys/vfs.h>
#include <sys/types.h>
#include <signal.h>

#include "shingledfs.h"
#include "emulator.h"
#include "helper.h"
#include "failstop.h"
#include "buffercache.h"
#include "log.h"

#include "inode.h"
#include "band_bitmap.h"
#include "band_log.h"
#include "cleaning_thread.h"
#include "band_meminfo.h"

shingledfs_data_t shingledfs_data;

/*******************************************************************************
 * FUSE INITIALIZATION AND CLEANUP FUNCTIONS
 ******************************************************************************/

/**
 * @brief Initializes the Shingled File System, preparing it for usage.
 *
 * @todo Buffer Cache size hard coded!
 *
 * @param conn FUSE structure detailing connection information
 */
shingledfs_data_t
*shingledfs_init(struct fuse_conn_info *conn) {
#ifdef GENERATE_LOGS
    char filename[50];
    int len;
    time_t curtime = time(NULL);

    /* Initialize the logging interface */
    len = strftime(filename, 50, "/tmp/smr%Y%m%d%H%M%S_main.log",
                   localtime(&curtime));
    if (!len) {
        fprintf(stderr, "Error generating log file name: %s\n", filename);
        assert(0);
    }
    log_open(filename);
#endif


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_init()\n");


    /* Load the Emulated Disk Image and its parameters */
    dod(!edi_init());
#ifdef _EDI_ON_RAW
    dod((SFSDATA.fd_edi = edi_open(SFSDATA.path_edi_metadata, SFSDATA.path_edi, &SFSDATA.fd_edi_metadata)) >= 0);
    read_edi_parameters(SFSDATA.fd_edi_metadata, &SFSDATA.edi_blksize,
                        &SFSDATA.edi_bandcount, &SFSDATA.edi_bandsize);
#else
    dod((SFSDATA.fd_edi = edi_open(SFSDATA.path_edi)) >= 0);
    read_edi_parameters(SFSDATA.fd_edi, &SFSDATA.edi_blksize,
                        &SFSDATA.edi_bandcount, &SFSDATA.edi_bandsize);
#endif

    /* Initialize the Buffer Cache */
    bc_init(1 << 30);


    /* Initialize the internal data structures and files */
    dod(!init_inodebitmap());
    //dod(!init_inodelist());

    /* Init bandmeminfo */
    dod(!init_band_meminfo());
    dod(!recover_band_log_data());

    log(LOG_SFS_INFO, "Initialized ShingledFS on-disk data structures!\n");

    /* Return the ShingledFS data structure for later use */
    log(LOG_SFS_RETURN, "ShingledFS initialized!\n");

    /* Initing the cleaner module. This will create a new thread for the cleaner */
    cleaner_init();

    return &shingledfs_data;
}

void shingledfs_destroy(void *data)
{
    int retval;

    log(LOG_SFS_CALLS, "shingledfs_destroy()\n");

    /* De-init buffer cache */
    retval = bc_destory();
    assert(retval == 0);

    CHECK_ERROR("bc_destory");
    /*
     * thread_join() for cleaner and buffer threads ?
     * We need to make sure that neither of these threads do anything now:
     * Options :
     * 1) Set a global and restart the threads just for them to start and die ?
     * 2) buffer_cache has already evicted everything so it should not run, just
     *    make sure cleaner thread is not doing anything and does not do
     *    anything after this ?
     * 3) Don't care for now
     */

    /* De-init band meminfo */
    cleanup_band_meminfo();
}

/*******************************************************************************
 * DIRECTORY MANAGEMENT FUNCTIONS
 ******************************************************************************/

/**
 * @brief Creates a new directory
 *
 * Documentation in fuse/fuse.h mentions that the mode bit for the directory may
 * not have been set. So it should be ORed with S_IFDIR to ensure that correct
 * directory mode is setup.
 *
 * @param path Path to the new directory
 * @param mode Permissions to set for the newly created directory
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_mkdir(const char     *path,
                 mode_t          mode) {
    char path_unshingled[PATH_MAX];

    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_mkdir(path=%s, mode=0%3o)\n", path, mode);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);

    /* Create the directory on the unshingled partition */
    dod(!mkdir(path_unshingled, mode | S_IFDIR));

    log(LOG_SFS_RETURN, "Created %s\n", path_unshingled);

    return 0;
}


/**
 * @brief Removes a directory
 *
 * @param path Path to the directory to be removed
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_rmdir(const char *path) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_rmdir(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);

    /* Remove the directory on the unshingled partition */
    if (rmdir(path_unshingled) < 0) {
        log(LOG_SFS_RETURN, "Remove of %s failed. errno=%d\n",
            path_unshingled, errno);
        return -errno;
    }


    log(LOG_SFS_RETURN, "Removed %s\n", path_unshingled);
    return 0;
}


/**
 * @brief Opens the specified directory
 *
 * @param path Path to the directory to open
 * @param fi FUSE File Info structure to store the Directory Pointer
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_opendir(const char                   *path,
                   struct fuse_file_info        *fi) {
    char path_unshingled[PATH_MAX];
    DIR *dp;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_opendir(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Open the directory */
    dod((dp = opendir(path_unshingled)) != NULL);

    /* Store the directory pointer in the FUSE info structure for future use */
    fi->fh = (intptr_t) dp;


    log(LOG_SFS_RETURN, "Opened directory %s (ptr %p)\n", path_unshingled, dp);
    return 0;
}



/**
 * @brief Reads the list of files in the specified directory
 *
 * This function filters out all ShingledFS internal files.
 *
 * @param path Path to the directory
 * @param buf Buffer to hold the list of files in the directory
 * @param fill_dir Pointer to function that will fill out the buffer
 * @param offset
 * @param file_info fuse_file_info structure from the call to opendir()
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_readdir(const char                   *path,
                   void                         *buf,
                   fuse_fill_dir_t              filler,
                   off_t                        offset,
                   struct fuse_file_info        *fi) {
    DIR *dp;
    struct dirent *de;
    int ent_ret, ent_total;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_readdir(path=%s, buf=%p, offset=%ld)\n",
        path, buf, offset);


    /* Grab the file handle for the directory from the FUSE info structure */
    dp = (DIR *) (uintptr_t) fi->fh;

    /* Read directory entries until possible */
    ent_ret = ent_total = 0;
    while ((de = readdir(dp)) != NULL) {
        /* Increment the total entries counter */
        ent_total++;

        /* Filter out ShingledFS internal files */
        if (strncmp(de->d_name, FILE_PREFIX, strlen(FILE_PREFIX))) {
            /* Add the entry to the buffer */
            if (filler(buf, de->d_name, NULL, 0) != 0) {
                log(LOG_SFS_ERRORS, "filler(): Buffer overflow!\n");
                return -EOVERFLOW;
            }

            /* Increment the returned entries counter */
            ent_ret++;
        }
    }


    log(LOG_SFS_RETURN, "Returned %d of %d entries\n", ent_ret, ent_total);
    return 0;
}



/**
 * @brief Closes the open directory
 *
 * @param path Path to the directory to close
 * @param fi fuse_file_info structure storing the Directory Pointer from opendir
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_releasedir(const char                *path,
                      struct fuse_file_info     *fi) {
    DIR *dp;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_releasedir(path=%s)\n", path);

#ifdef GENERATE_LOGS
    /* Translate the path to the Unshingled partition */
    char path_unshingled[PATH_MAX];
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);
#endif


    /* Grab the file handle for the directory from the FUSE info structure */
    dp = (DIR *) (uintptr_t) fi->fh;

    /* Close the open directory */
    if (closedir(dp)) {
        log(LOG_SFS_ERRORS, "closedir(%s): %s\n",
            path_unshingled, strerror(errno));
        return -errno;
    }


    log(LOG_SFS_RETURN, "Closed directory %s (ptr %p)\n", path_unshingled, dp);
    return 0;
}



/*******************************************************************************
 * FILE MANAGEMENT FUNCTIONS
 ******************************************************************************/

/**
 * @brief Creates a new file node.
 *
 * This function calls mknod() to create the requested file along with its
 * Attribute Replication file, into which the attributes of the file are stored.
 *
 * This function is called by FUSE to create non-directory non-symlink nodes. If
 * create() is defined, then that is called. However, create() only works in
 * Linux kernel after 2.6.15. So it was better to implement the mknod() function
 * instead of create().
 *
 * @param path Path to the file to create
 * @param mode Access mode specifications for the newly created file
 * @param file_info fuse_file_info structure from the call to opendir()
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_mknod(const char     *path,
                 mode_t          mode,
                 dev_t           dev) {
    char path_unshingled[PATH_MAX];
    char path_xattr[PATH_MAX];
    struct stat statbuf;
    mode_t xattr_mode = S_IFREG|S_IWUSR|S_IRUSR|S_IWGRP|S_IRGRP|S_IWOTH|S_IROTH;
    inode_t inode;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_mknod(path=%s, mode=0%3o, dev=%lu)\n", path, mode, dev);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);

    /* Generate the name of the Attribute Replication file */
    xattr_path(path_unshingled, path_xattr);


    /* Create the file */
    dod(!mknod(path_unshingled, mode, dev));

    /* Read the attributes of the newly created file */
    dod(!stat(path_unshingled, &statbuf));


    /* Create the Attribute Replication file */
    dod(!mknod(path_xattr, xattr_mode, dev));


    /* Initialize the inode structure and allocate one for the new file */
    memset(&inode, 0, sizeof(inode_t));
    dod(inode.inode_num = get_freeinodenum());


    /* Setup extended attributes for the first time. */
    dod(!lsetxattr(path_xattr, X_STATS, &statbuf, sizeof(struct stat), 0));
    xattr_set(path_unshingled, X_INODE, &inode.inode_num,
              sizeof(inode.inode_num), 0);

    /* Add the inode_xattr file */
    char inode_xattr_path[255];
    xattr_path_from_inode(SFSDATA.path_unshingled, inode_xattr_path, inode.inode_num);
    dod(!mknod(inode_xattr_path, xattr_mode, dev));

    /* Add the inode to the inode file */
    dod(!put_inode(inode.inode_num, &inode));


    log(LOG_SFS_RETURN, "Created %s (inode %d), %s and replicated attributes\n",
        path_unshingled, inode.inode_num, path_xattr);
    return 0;
}



/**
 * @brief Opens the specified file.
 *
 * Essentially, this function just calls @see bc_open() to load the file into
 * the buffer cache, and stores the @see bc_file_t record of the newly opened
 * file.
 *
 * @param path Path to the file to create
 * @param file_info fuse_file_info structure to store file-related data
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_open(const char              *path,
                struct fuse_file_info   *fi) {
    char path_unshingled[PATH_MAX];
    bc_file_t *bc_file;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_open(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);

    /* Pass the task of opening the file to the buffer cache */
    bc_file = bc_open(path_unshingled);

    /* Save a reference to the BC File record for later use */
    fi->fh = (uint64_t)bc_file;


    log(LOG_SFS_RETURN, "Opened %s (fd = %d)\n", path_unshingled, bc_file->fd);
    return 0;
}



/**
 * @brief Reads data from an open file.
 *
 * @param path Path to the open file.
 * @param buf Buffer to hold the data read from the file.
 * @param size Size of the data to read from the file (in bytes)
 * @param offset Offset into the file from where the data is to be read.
 * @param file_info FUSE structure created when the file was opened.
 *
 * @return The no. of bytes read, on success; 0 on EOF; -1 on failure
 */
int
shingledfs_read(const char              *path,
                char                    *buf,
                size_t                  size,
                off_t                   offset,
                struct fuse_file_info   *fi) {
    bc_file_t *bc_file;
    int bytes_read;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_read(path=%s, buf=%p, size=%ld, offset=%ld)"
        "\n", path, buf, size, offset);


    /* Read data from the file */
    bc_file = (bc_file_t *)(fi->fh);
    if ((bytes_read = pread(bc_file->fd, buf, size, offset)) < 0) {
        log(LOG_SFS_RETURN | LOG_SFS_ERRORS, "pread(): %s\n", strerror(errno));
        return -errno;
    }


    log(LOG_SFS_RETURN, "Read %d bytes\n", bytes_read);
    return bytes_read;
}



/**
 * @brief Write data to an open file.
 *
 * If the write() fails, this function checks to see if the device is full; i.e.
 * tmpFS is out of space. If this is the case, then @see bc_clean() is called to
 * evict a file to create more space in tmpFS.
 *
 * @param path Path to the open file.
 * @param buf Buffer holding the data to be written.
 * @param size Size of the data to be written (in bytes).
 * @param offset Offset into the file where the data is to be written.
 * @param file_info FUSE structure created when the file was opened.
 *
 * @return The no. of bytes written, on success; -1 on failure
 */
int
shingledfs_write(const char             *path,
                 const char             *buf,
                 size_t                 size,
                 off_t                  offset,
                 struct fuse_file_info  *fi) {
    bc_file_t *bc_file;
    ssize_t bytes_written;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_write(path=%s, buf=%p, size=%ld, offset=%ld)"
        "\n", path, buf, size, offset);


    /* Retrieve the BC File record for the file */
    bc_file = (bc_file_t *)(fi->fh);


    /* Write data to the file */
    while ((bytes_written = pwrite(bc_file->fd, buf, size, offset)) < 0) {
        /* Check if the error is because the buffer cache is out of space */
        if (errno != ENOSPC) {
            log(LOG_SFS_RETURN | LOG_SFS_ERRORS, "pwrite(): %s\n",
                strerror(errno));
            return -errno;
        }

        /* Clean the buffer cache */
        bc_clean();
    }

    /* Update the file stats */
    dod(!pthread_mutex_lock(&(bc_file->mutex)));
    bc_file->dirty = 1;
    dod(!pthread_mutex_unlock(&(bc_file->mutex)));


    log(LOG_SFS_RETURN, "Wrote %d bytes to fd %d\n", bytes_written, bc_file->fd);
    return bytes_written;
}



/**
 * @brief Flushes file contents to disk.
 *
 * For ShingledFS, this function does nothing! We want to write file contents to
 * disk on close(), not before that!
 *
 * @param path Path to the open file.
 * @param file_info FUSE structure created when the file was opened.
 *
 * @return 0, to indicate success
 */
int
shingledfs_flush(const char             *path,
                 struct fuse_file_info  *fi) {
    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_flush(path=%s)\n", path);


    log(LOG_SFS_RETURN, "\n", path);
    return 0;
}



/**
 * @brief Closes an open file
 *
 * @param path Path to the file to close
 * @param file_info fuse_file_info structure to store file-related data
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_release(const char                   *path,
                   struct fuse_file_info        *fi) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_release(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Close the file in the buffer cache */
    bc_close((bc_file_t *)(fi->fh));


    log(LOG_SFS_RETURN, "Successfully closed %s\n", path);
    return 0;
}



/**
 * @brief Renames a file
 *
 * @param path Path to the file to be renamed
 * @param path_new Path to the renamed file
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_rename(const char    *path,
                  const char    *path_new) {
    char path_unshingled[PATH_MAX], path_new_unshingled[PATH_MAX];
    char path_xattr[PATH_MAX], path_new_xattr[PATH_MAX];
    char path_bc[PATH_MAX];
    struct stat statbuf;
    int inode_num;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_rename(path=%s, path_new=%s)\n",
        path, path_new);

    /* Translate the paths */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);
    translate_path(path_new, SFSDATA.path_unshingled, path_new_unshingled);
    xattr_path(path_unshingled, path_xattr);
    xattr_path(path_new_unshingled, path_new_xattr);


    /* Read the attributes of the file */
    dod(!lstat(path_unshingled, &statbuf));
    dod(!xattr_get(path_unshingled, X_INODE, &inode_num, sizeof(int)));

    /* Check if the path is to a symlink pointing to the tmpFS file. */
    if (S_ISLNK(statbuf.st_mode)) {
        /* Rename the tmpFS copy of the file */
        if (bc_rename(path_unshingled, path_new_unshingled))
            log(LOG_SFS_INFO, "%s was not found in the buffer cache!\n",
                path_unshingled);

        /* Generate the tmpFS file path */
        bc_generate_tmpFS_path(path_bc, path_new_unshingled, inode_num);

        /* Recreate the symbolic link with the new name */
        while (symlink(path_bc, path_new_unshingled)) {
            /* Check if the path already exists. */
            if (errno != EEXIST) {
                log(LOG_SFS_ERRORS, "Error creating symbolic link from %s to %s"
                    ": %s\n", path_new_unshingled, path_bc,
                    strerror(errno));
                dod(0);
            }

            /* If the path already exists, then delete it and create the symlink
             * because rename() atomically replaces the target. We need the same
             * behavior. */
            dod(!unlink(path_new_unshingled));
        }

        /* Delete the old symlink */
        dod(!unlink(path_unshingled));

    } else {
        /* Rename the file */
        dod(!rename(path_unshingled, path_new_unshingled));
    }


    /* Rename the Attribute Replication file */
    dod(!rename(path_xattr, path_new_xattr));


    log(LOG_SFS_RETURN, "File rename successful!\n");
    return 0;
}



/**
 * @brief Deletes a file
 *
 * This function deletes the file and its attribute replication file, if any.
 *
 * @param path Path to the file to be deleted
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_unlink(const char *path) {
    char path_unshingled[PATH_MAX], path_xattr[PATH_MAX];
    struct stat statbuf;
    inode_t inode;
    int inodeno;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_unlink(path=%s)\n", path);

    /* Translate the paths */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);
    xattr_path(path_unshingled, path_xattr);


    /* Read the attributes of the file */
    dod(!lstat(path_unshingled, &statbuf));


    /* Check if the path is to a symlink pointing to the tmpFS file. */
    if (S_ISLNK(statbuf.st_mode)) {
        if (!bc_unlink(path_unshingled)) {
            log(LOG_SFS_INFO, "Unlinked %s in tmpFS!\n", path_unshingled);
        }

        /* bc_unlink() fails when the file is not present in tmpFS. This can
         * happen when the file has been evicted from tmpFS, but is still
         * present in the shingled disk */
    }

    /* Read the inode information for the file */
    if (!xattr_get(path_unshingled, X_INODE, &inodeno, sizeof(int))) {
        /*
         * xattr_get will succeed only if we had created this link, if the link
         * was created by the system, unlink it directly out the if {}
         */
        dod(!get_inode(inodeno, &inode));

        /* Unlink the inode */
        while(inode_unlink(inodeno, inode) != 0)
            dod(!get_inode(inodeno, &inode));

        inode_xattr_free(inodeno);
    }

    /* Delete the file */
    dod(!unlink(path_unshingled));
    log(LOG_SFS_INFO, "Deleted %s\n", path_unshingled);

    /* Delete the Attribute Replication file. If the operation fails, ensure
     * that it is NOT ENOENT. This can be the case when deleting symlinks. */
    if (unlink(path_xattr)) {
        if (errno != ENOENT) {
            log(LOG_SFS_ERRORS, "unlink(%s): %s\n",
                path_xattr, strerror(errno));
            dod(0);
        }
    } else {
        log(LOG_SFS_INFO, "Deleted %s\n", path_xattr);
    }


    log(LOG_SFS_RETURN, "File deletion successful!\n");
    return 0;
}



/**
 * @brief Truncates the file at the specified offset
 *
 * @todo This function does not need to update the inode information as that
 * will result in leaking of data blocks. Instead, ensure that when a file is
 * loaded, the number of blocks is calculated based on its actual file size; NOT
 * the no. of blocks saved in the inode.
 *
 * @param path Path to the file to be truncated
 * @param offset Offset into the file to truncate at
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_truncate(const char       *path,
                    off_t            offset) {
    char path_unshingled[PATH_MAX];
    struct stat statbuf;
    int inode_no;
    inode_t inode;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_truncate(path=%s, offset=%ld)\n",
        path, offset);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the attributes of the file */
    dod(!lstat(path_unshingled, &statbuf));


    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf.st_mode)) {
        /* Update the size in the Attribute Replication file to the specified
         * offset */
        dod(!xattr_get(path_unshingled, X_STATS, &statbuf,
                       sizeof(struct stat)));
        statbuf.st_size = offset;
        xattr_set(path_unshingled, X_STATS, &statbuf, sizeof(struct stat), 0);

    } else if (S_ISLNK(statbuf.st_mode)) {
        /* Truncate the file to the specified size */
        dod(!truncate(path_unshingled, offset));
    }


    /* Update the file's inode */
    dod(!xattr_get(path_unshingled, X_INODE, &inode_no, sizeof(int)));

    lock_inodelist();
    dod(!get_inode(inode_no, &inode));

    //inode.size = (offset + SFSDATA.edi_blksize - 1) / SFSDATA.edi_blksize;
    //dod(!put_inode(inode_no, &inode));

    size_t size = (offset + SFSDATA.edi_blksize - 1) / SFSDATA.edi_blksize;
    inode_update(&inode, inode.band_id, inode.band_offset, size);

    unlock_inodelist();

    return 0;
}

int shingledfs_fsync(const char * path, int datasync, struct fuse_file_info *fi) {
    /*
     * Here we have to close the file for it to get enqueued in the thread in buffercache
     * to be written to disk.
     */
    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_fsync(path=%s,datasync=%d)\n", path);
 
    /* Wait till file is flushed to disk (if it is present in the buffercache queue */
    bc_block_until_file_is_processed((bc_file_t *)(fi->fh));

    log(LOG_SFS_RETURN, "Successfully fsynced %s\n", path);

    return 0;
}

///**
// * @brief Truncates the file at the specified offset
// *
// * This function relies on the file descriptor of an open file, passed in the
// * fuse_file_info structure.
// *
// * @param path Path to the file whose attributes are to be retrived
// * @param offset Offset into the file to truncate at
// * @param file_info fuse_file_info structure from the call to open() or create()
// *
// * @return 0 on success; -errno on failure
// */
//int shingledfs_ftruncate(const char *path, off_t offset,
//              struct fuse_file_info *fi) {
//    char path_unshingled[PATH_MAX];
//    int inode_no, fd_log;
//    inode_t inode;
//
//
//    /* Log the call */
//      log(LOG_SFS_CALLS, "shingledfs_ftruncate(path=%s, offset=%ld)\n", path, offset);
//
//    /* Translate the path to the Unshingled partition */
//    if (translate_path(path, SFSDATA.path_unshingled, path_unshingled)) {
//      log("translate_path(%s): %s\n", path, strerror(errno));
//      assert(0);
//      return -errno;
//      }
//
//
//    /* Truncate the file */
//    if (ftruncate(((bc_file_t *)(fi->fh))->fd, offset)) {
//      log("ftruncate(%s, %ld): %s\n", path, offset, strerror(errno));
//      assert(0);
//      return -errno;
//      }
//
//
//    /* Update the file's inode */
//      assert(!xattr_get(path_unshingled, X_INODE, &inode_no, sizeof(int)));
//      assert(!get_inode(inode_no, &inode));
//      inode.size = (offset + SFSDATA.edi_blksize - 1) / SFSDATA.edi_blksize;
//      assert(!put_inode(inode_no, &inode));
//
//
//      /* Write the delete log */
//      assert(offset == 0);
//      fd_log = init_deletelog(inode.band_id);
//      assert(fd_log);
//      append_deletelog(fd_log, inode.band_offset, inode.size, inode.entrynumber);
//      close(fd_log);
//      log("Delete Log entry: Band %ld, RBA %ld, %d blocks, "
//                      "Append Log Entry: %d!\n", inode.band_id, inode.band_offset,
//                      inode.size, inode.entrynumber);
//
//    return 0;
//}



/**
 * @brief Retrieves the attributes for the specified file.
 *
 * @todo Is it cheaper to perform a comparison on with every open file in the BC
 * to determine location? The "file is open" code path takes 3 system calls!
 *
 * @param path Path to the file whose attributes are to be retrived
 * @param statbuf Buffer to hold the retrieved stats
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_getattr(const char   *path,
                   struct stat  *statbuf) {
    char path_unshingled[PATH_MAX], path_target[PATH_MAX];
    int len;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_getattr(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the attributes of the file */
    if (lstat(path_unshingled, statbuf)) {
        log(LOG_SFS_RETURN | LOG_SFS_ERRORS , "lstat(): %s\n", strerror(errno));
        return -errno;
    }


    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf->st_mode)) {
        /* Read the stats for the file from the Attribute Replication file */
        dod(!xattr_get(path_unshingled, X_STATS, statbuf, sizeof(struct stat)));

    } else if (S_ISLNK(statbuf->st_mode)) {
        /* Check if the symlink is a ShingledFS internal symlink i.e. it points
         * to a file on tmpFS */
        dod(readlink(path_unshingled, path_target, PATH_MAX) > 0);

        /* If the symlink is a ShingledFS internal symlink, then read the stats
         * from the tmpFS file */
        len = strlen(SFSDATA.path_tmpfs);
        if (!strncmp(path_target, SFSDATA.path_tmpfs, len)) {
            dod(!stat(path_unshingled, statbuf));
        }
    }


    log(LOG_SFS_RETURN, "File stats returned successfully!\n");
    return 0;
}



/**
 * @brief Retrieves the attributes for the specified file.
 *
 * This function relies on the file descriptor of an open file, passed in the
 * fuse_file_info structure.
 *
 * As per documentation in fuse/fuse.h, this is currently only called after
 * create(), but may later be used for invocations of fstat() as well.
 *
 * @param path Path to the file whose attributes are to be retrived
 * @param statbuf Buffer to hold the retrieved stats
 * @param file_info fuse_file_info structure from the call to open()
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_fgetattr(const char                          *path,
                    struct stat                         *statbuf,
                    struct fuse_file_info       *fi) {
    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_fgetattr(path=%s)\n", path);

    /* Read the stats for the file */
    dod(!fstat(((bc_file_t *)(fi->fh))->fd, statbuf));


    log(LOG_SFS_RETURN, "File stats returned successfully!\n");
    return 0;
}



/**
 * @brief Sets the extended attributes
 *
 * @param path Path to the file whose extended attribute is to be set
 * @param name Name of the extended attribute to update
 * @param value Buffer holding the value to be set for the extended attribute
 * @param size Size of the data in the buffer
 * @param flags Flags to refine operation semantics
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_setxattr(const char *path,
                    const char *name,
                    const char *value,
                    size_t size,
                    int flags) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_setxattr(path=%s, name=%s, value=%p, "
        "size=%ld, flags=%d)\n", path, name, value, size, flags);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Ensure that there are no naming clashes with ShingledFS internals */
    dod((!strcmp(name, X_INODE)) || (!strcmp(name, X_STATS)));

    /* Set the extended attributes on the Attribute Replication file */
    xattr_set(path_unshingled, name, (void *)value, size, flags);


    log(LOG_SFS_RETURN, "Set extended attribute %s to %ld-byte value at %p on "
        "Attribute Replication file\n", name, size, value);
    return 0;
}



/**
 * @brief Retrieves the value associated with an extended attribute
 *
 * @param path Path to the file whose extended attribute are to be retrieved
 * @param name Name of the extended attribute to retrieve
 * @param value Buffer to hold the value of the extended attribute
 * @param size Size of the data in the buffer
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_getxattr(const char  *path,
                    const char  *name,
                    char        *value,
                    size_t      size) {
    char path_unshingled[PATH_MAX];

    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_getxattr(path=%s, name=%s, value=%p, "
        "size=%ld)\n", path, name, value, size);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);

    if(is_valid_dir(path_unshingled)) {
        //FIXME: Xattr for directory not supported. Return if trying to getxattr on directory.
        return 0;
    }

    /* Ensure that there are no naming clashes with ShingledFS internals */
    dod((strcmp(name, X_INODE)) && (strcmp(name, X_STATS)));

    /* Get the extended attributes */
    if (xattr_get(path_unshingled, name, value, size)) {
        log(LOG_SFS_RETURN | LOG_SFS_ERRORS, "xattr_get(): %s\n",
            strerror(errno));
        return -errno;
    }


    log(LOG_SFS_RETURN, "Retrieved a %ld-byte value at %p for extended "
        "attribute %s from Attribute Replication file\n",
        size, value, name);
    return 0;
}



/**
 * @brief Retrieves the list of extended attributes
 *
 * @param path Path to the file whose extended attributes are to be set
 * @param list Buffer to hold the list of extended attributes
 * @param size Size of the data in the buffer
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_listxattr(const char    *path,
                     char          *list,
                     size_t        size) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_listxattr(path=%s, list=%p)\n", path, list);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Get the list of extended attributes */
    dod(!llistxattr(path_unshingled, list, size));


    log(LOG_SFS_RETURN, "Listed %ld extended attributes from Attribute "
        "Replication file into list at %p\n", size, list);
    return 0;
}



/**
 * @brief Removes an extended attribute
 *
 * @param path Path to the file whose extended attributes are to be removed
 * @param name Name of the extended attribute to update
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_removexattr(const char       *path,
                       const char       *name) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_removexattr(path=%s, name=%s)\n",
        path, name);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Remove the extended attribute */
    dod(!lremovexattr(path_unshingled, name));


    log(LOG_SFS_RETURN, "Removed extended attribute %s from Attribute "
        "Replication file\n", name);
    return 0;
}



/**
 * @brief Returns the access mask for the specified path
 *
 * @todo Implement access() correctly when the file is not open. This should
 * involve check the uid and gid of the file and the current user and the access
 * permissions set in st_mode in the file's attributes.
 *
 * @param path Path to the file whose access permissions are to be checked
 * @param mask Permissions to check for
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_access(const char       *path,
                  int               mask) {
    char path_unshingled[PATH_MAX];
    struct stat stats, *statbuf;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_access(path=%s, mask=0%3o)\n", path, mask);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the attributes of the file */
    statbuf = &stats;
    dod(!lstat(path_unshingled, statbuf));


    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf->st_mode)) {
        /* Read the stats for the file from the Attribute Replication file */
        dod(!xattr_get(path_unshingled, X_STATS, statbuf, sizeof(struct stat)));

        /* Check if the request is for file existence */
        if (mask == F_OK)
            return 0;

        /* Check the access credentials */
        if (!((statbuf->st_mode) & (mask << 6))) {
            log(LOG_SFS_RETURN | LOG_SFS_ERRORS, "File permissions 0%3o don't "
                "permit 0%3o!\n",
                statbuf->st_mode, mask);
            errno = EPERM;
            return -1;
        }

    } else {
        /* Check the access permissions */
        dod(!access(path_unshingled, mask));
    }


    log(LOG_SFS_RETURN, "Access permission granted!\n");
    return 0;
}



/**
 * @brief Sets the access and/or modification times of the specified file
 *
 * @todo Update to account for attribute replication
 *
 * @param path Path to the file to create
 * @tv utimebuf structure containing the new timestamps
 *
 * @return 0 on success; -errno on failure
 */
int
shingledfs_utime(const char     *path,
                 struct utimbuf *tv) {
    char path_unshingled[PATH_MAX];
    struct stat stats, *statbuf;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_utime(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the attributes of the file */
    statbuf = &stats;
    dod(!lstat(path_unshingled, statbuf));


    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf->st_mode)) {
        /* Update the time stamps */
        dod(!xattr_get(path_unshingled, X_STATS, statbuf, sizeof(struct stat)));
        statbuf->st_atime = tv->actime;
        statbuf->st_mtime = tv->modtime;
        xattr_set(path_unshingled, X_STATS, statbuf, sizeof(struct stat), 0);

    } else {
        /* Set the access and/or modification times of the file */
        dod(!utime(path_unshingled, tv));
    }


    log(LOG_SFS_RETURN, "File timestamps updated successfully!\n");
    return 0;
}



/**
 * @brief Changes the permission bits of a file
 *
 * @param path Path to the file whose permission bits are to be changed
 * @param mode The new permission bits to be set
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_chmod(const char     *path,
                 mode_t          mode) {
    char path_unshingled[PATH_MAX];
    struct stat stats, *statbuf;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_chmod(path=%s, mode=0%o)\n", path, mode);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the attributes of the file */
    statbuf = &stats;
    dod(!lstat(path_unshingled, statbuf));

    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf->st_mode)) {
        /* Update the permissions */
        dod(!xattr_get(path_unshingled, X_STATS, statbuf, sizeof(struct stat)));
        statbuf->st_mode = mode;
        xattr_set(path_unshingled, X_STATS, statbuf, sizeof(struct stat), 0);

    } else {
        /* Change the permission bits of the file */
        dod(!chmod(path_unshingled, mode));
    }


    log(LOG_SFS_RETURN, "File access permissions updated successfully!\n");
    return 0;
}



/**
 * @brief Changes the owner and group of a file
 *
 * @todo Fix this function to account for Attribute Replication
 *
 * @param path Path to the file whose owner and group is to be changed
 * @param uid The UID of the user who is the new owner of the file
 * @param gid The GID of the group that can access the file
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_chown(const char     *path,
                 uid_t           uid,
                 gid_t           gid) {
    char path_unshingled[PATH_MAX];
    struct stat stats, *statbuf;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_chown(path=%s, uid=%d, gid=%d)\n",
        path, uid, gid);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the attributes of the file */
    statbuf = &stats;
    dod(!lstat(path_unshingled, statbuf));


    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf->st_mode)) {
        /* Update the file owners */
        dod(!xattr_get(path_unshingled, X_STATS, statbuf, sizeof(struct stat)));
        statbuf->st_uid = uid;
        statbuf->st_gid = gid;
        xattr_set(path_unshingled, X_STATS, statbuf, sizeof(struct stat), 0);

    } else {
        /* Change the uid of the file */
        dod(!chown(path_unshingled, uid, gid));
    }


    log(LOG_SFS_RETURN, "File owner and/or group updated successfully!\n");
    return 0;
}



/**
 * @brief Creates a symbolic link
 *
 * @param path Path to the symbolic link to be created
 * @param path_target Path to the target of the symbolic link
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_symlink(const char   *path,
                   const char   *path_target) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_symlink(path=%s, path_target=%s)\n",
        path, path_target);

    /* Translate the link target path to the Unshingled partition */
    translate_path(path_target, SFSDATA.path_unshingled, path_unshingled);


    /* Create the symbolic link */
    dod(!symlink(path, path_unshingled));


    log(LOG_SFS_RETURN, "Created symlink from %s to %s\n", path_unshingled, path);
    return 0;
}



/**
 * @brief Reads the target of a symbolic link
 *
 * @param path Path to the symbolic link whose target is to be determined
 * @param buf Buffer to hold the target path
 * @param size Size of the buffer
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_readlink(const char    *path,
                    char          *buf,
                    size_t         size) {
    char path_unshingled[PATH_MAX];
    int bytes_written;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_readlink(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the target of the link */
    dod((bytes_written = readlink(path_unshingled, buf, size)) >= 0);

    /* Null-terminate the buffer */
    if (bytes_written == size)
        bytes_written--;
    buf[bytes_written] = '\0';


    log(LOG_SFS_RETURN, "Symlink %s points to %s\n", path_unshingled, buf);
    return 0;
}



/**
 * @brief Creates a hard link
 *
 * @todo This function might not be suited for the ShingledFS architecture.
 * Consider what happens to a hard link when the target is opened. The file is
 * deleted and replaced with a symlink to a file on tmpFS. With the deletion,
 * the hard link is removed. Unless the hard links are tracked and recreated on
 * each open and BC eviction, this is bound to fail.
 *
 * @param path Path to the hard link to be created
 * @param path_target Path to the target of the symbolic link
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_link(const char      *path,
                const char      *path_target) {
    char path_unshingled[PATH_MAX], path_target_unshingled[PATH_MAX];
    struct stat stats, *statbuf;


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_link(path=%s, path_target=%s)\n",
        path, path_target);

    /* Translate the paths */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);
    translate_path(path_target, SFSDATA.path_unshingled,
                   path_target_unshingled);


    /* Create the link */
    dod(!link(path_unshingled, path_target_unshingled));


    /* Read the attributes of the file */
    statbuf = &stats;
    dod(lstat(path_unshingled, statbuf));

    /* Check if the path is to a regular file. */
    if (S_ISREG(statbuf->st_mode)) {
        /* Read the attributes of the file */
        statbuf = &stats;
        /* Update the no. of hard links */
        dod(!xattr_get(path_unshingled, X_STATS, statbuf, sizeof(struct stat)));
        statbuf->st_nlink++;
        xattr_set(path_unshingled, X_STATS, statbuf, sizeof(struct stat), 0);
    }


    log(LOG_SFS_RETURN, "Created link from %s to %s\n",
        path_unshingled, path_target_unshingled);
    return 0;
}



/**
 * @brief Returns ShingledFS file system statistics
 *
 * @param path Path to any file on ShingledFS
 * @param statv Buffer to hold the statistics returned
 *
 * @return 0 on success; -1 on failure
 */
int
shingledfs_statfs(const char            *path,
                  struct statvfs        *statv) {
    char path_unshingled[PATH_MAX];


    /* Log the call */
    log(LOG_SFS_CALLS, "shingledfs_statfs(path=%s)\n", path);

    /* Translate the path to the Unshingled partition */
    translate_path(path, SFSDATA.path_unshingled, path_unshingled);


    /* Read the stats for the local file system */
    dod(!statvfs(path_unshingled, statv));


    log(LOG_SFS_RETURN, "bsize: %ld, frsize: %ld, blocks: %ld, bfree: %ld, "
        "bavail: %ld, files: %ld, ffree: %ld, favail: %ld, "
        "fsid: %ld, namemax: %d\n",
        statv->f_bsize, statv->f_frsize, statv->f_blocks,
        statv->f_bfree, statv->f_bavail, statv->f_files,
        statv->f_ffree, statv->f_favail, statv->f_fsid,
        statv->f_namemax);
    return 0;
}
