/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file helper.c
 *
 * @brief Implementation of Helper functions
 *
 * @author Anand Suresh <anandsur@andrew.cmu.edu>
 */

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/xattr.h>
#include <linux/limits.h>

#include "helper.h"
#include "failstop.h"
#include "emulator.h"
#include "band_log.h"
#include "band_meminfo.h"


/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */



/* ====================== */
/* -- Static Variables -- */
/* ====================== */



/* ============================== */
/* -- Function Implementations -- */
/* ============================== */

/**
 * @brief Reads the EDI Parameters and saves them in the FUSE private context.
 *
 * @param fd_edi File Descriptor to open EDI file
 * @param blk_size Pointer to variable storing the block size of the EDI
 * @param band_count Pointer to variable storing the no. of bands on the EDI
 * @param band_size Pointer to variable storing the size of a band on the EDI
 */
void
read_edi_parameters(int                  fd,
                    uint32_t            *blk_size,
                    band_t              *band_count,
                    rba_t               *band_size) {
    /* Read the EDI Block Size */
    dod(edi_modesense(fd, INQ_BLKSIZE, blk_size) == 0);

    /* Read the EDI Band Size */
    dod(edi_modesense(fd, INQ_BNDSIZE, band_size) == 0);

    /* Read the EDI Band Count */
    dod(edi_modesense(fd, INQ_BNDCOUNT, band_count) == 0);
}



/**
 * @brief Translate the incoming path by adding the specified prefix to it.
 *
 * out_path must be large enough to store the concatenation of in_path and
 * prefix, plus the terminating null character.
 *
 * @param in_path The input path
 * @param prefix The path to prefix to the incoming path
 * @param out_path The location to store the resulting concatenation
 */
void
translate_path(const char       *in_path,
               const char       *prefix,
               char             *out_path) {
    /* Validate the inputs */
    if ((in_path == NULL) || (out_path == NULL)) {
        log(LOG_SFS_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }

    /* Ensure that the lengths of the paths will not be an issue */
    if (strlen(in_path) + strlen(prefix) >= PATH_MAX) {
        log(LOG_SFS_ERRORS, "Translated path exceeds PATH_MAX!\n");
        exit(EXIT_FAILURE);
    }

    /* Concatenate the prefix and in_path and store the result in out_path */
    strcpy(out_path, prefix);
    //strcat(out_path, "/");
    strcat(out_path, in_path);
}



/**
 * @brief Generates the name of the attribute replication file for the specified
 * input file path.
 *
 * This function should not be called for directories!
 *
 * Given the path to the input file, this function appends the Attribute
 * Replication file prefix (.xattr_) to the name of the file.
 *
 * @todo Consider adding the S_ISREG() check on the stats of the input path!
 *
 * @param in_path The path to the file on the unshingled partition
 * @param out_path The location to store the resulting concatenation
 */
void
xattr_path(const char   *in_path,
           char         *out_path) {
    char *filename;
    int len;


    /* Validate the inputs */
    if ((in_path == NULL) || (out_path == NULL)) {
        log(LOG_SFS_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }

    /* Ensure that the lengths of the paths will not be an issue */
    if (strlen(in_path) + strlen(FILE_XATTR_PREFIX) >= PATH_MAX) {
        log(LOG_SFS_ERRORS, "Translated path exceeds PATH_MAX!\n");
        exit(EXIT_FAILURE);
    }


    /* Locate the start of the filename in the input path */
    if ((filename = strrchr(in_path, '/')) == NULL) {
        log(LOG_SFS_ERRORS, "Received malformed path %s\n", in_path);
        exit(EXIT_FAILURE);
    }
    filename++;


    /* Construct the name of the attribute replication file */
    len = (size_t)filename - (size_t)in_path;
    strncpy(out_path, in_path, len);
    out_path[len] = '\0';
    strcat(out_path, FILE_XATTR_PREFIX);
    strcat(out_path, filename);
}


/**
 * @brief This function creates a .smr_xattr_<inode_num> file in the
 * unshingled space. This is the file that stores the inode struct in
 * its xattr. This function is needed because the cleaner needs the inode
 * structures based on only the inode numbers.
 */
void
xattr_path_from_inode(const char *in_path, char *out_path, int inode_num) {
    /* Validate the inputs */
    if ((in_path == NULL) || (out_path == NULL)) {
        log(LOG_SFS_ERRORS, "Null arguments passed!\n");
        exit(EXIT_FAILURE);
    }

    /* Ensure that the lengths of the paths will not be an issue */
    if ((strlen(in_path) + strlen(FILE_INODE_XATTR_PREFIX) + sizeof(inode_num) + 1) >= PATH_MAX) {
        log(LOG_SFS_ERRORS, "Translated path exceeds PATH_MAX!\n");
        exit(EXIT_FAILURE);
    }

    /* Construct the name of the inode based attribute replication file */
    sprintf(out_path, "%s/%s%d", in_path, FILE_INODE_XATTR_PREFIX, inode_num);
}


/**
 * @brief Retrieves the extended attributes of the specified file
 *
 * @param path The path to the file on the unshingled partition
 * @param name Name of the extended attribute to retrieve
 * @param value Buffer to hold the value of the extended attribute
 * @param size Size of the buffer
 *
 * @return 0 on success; -1 on failure
 */
int
xattr_get(const char    *path,
          const char    *name,
          void          *value,
          size_t         size) {
    char path_xattr[PATH_MAX];


    /* Generate the name of the Attribute Replication file */
    xattr_path(path, path_xattr);

    /* Get the attributes */
    return (lgetxattr(path_xattr, name, value, size) > 0) ? 0 : -1;
}



/**
 * @brief Sets the replicated attributes of the specified file
 *
 * @param path The path to the file on the unshingled partition
 * @param name Name of the extended attribute to update
 * @param value Buffer holding the value to be set for the extended attribute
 * @param size Size of the data in the buffer
 */
void
xattr_set(const char    *path,
          const char    *name,
          void          *value,
          size_t         size,
          int            flags) {
    char path_xattr[PATH_MAX];
    struct stat statbuf;


    /* Generate the name of the Attribute Replication file */
    xattr_path(path, path_xattr);

    /* Check if the Replicated Attributes are being updated */
    if (!strcmp(name, X_STATS)) {
        /* Read the Replicated attributes */
        dod(lgetxattr(path_xattr, X_STATS, &statbuf, sizeof(statbuf)) >= 0);

        /* Update the attributes */
        ((struct stat *)value)->st_dev = statbuf.st_dev;
        ((struct stat *)value)->st_ino = statbuf.st_ino;
        ((struct stat *)value)->st_rdev = statbuf.st_rdev;
        log_stats(path, value);
    }

    /* Set the attributes */
    dod(!lsetxattr(path_xattr, name, value, size, flags));
}



/**
 * @brief Validates the specified path to be a directory
 *
 * @param path The path to the directory to be verified
 *
 * @return 1 on successful validation; 0 on validation failure.
 */
int
is_valid_dir(const char *path) {
    struct stat fstats;

    /* Validate the path */
    if (stat(path, &fstats))
        return 0;

    if (!S_ISDIR(fstats.st_mode))
        return 0;

    return 1;
}



/**
 * @brief Validates the specified path to be a file
 *
 * @param path The path to the file to be verified
 *
 * @return 1 on successful validation; 0 on validation failure.
 */
int
is_valid_file(const char *path) {
    struct stat fstats;

    /* Validate the path */
    if (stat(path, &fstats))
        return 0;

    if (!S_ISREG(fstats.st_mode))
        return 0;

    return 1;
}



/**
 * @brief Performs all the updates needed to mark the band space defined by the
 * specified inode for cleaning.
 *
 * This function should be called whenever a space allocated to a file on any
 * band needs to be marked for cleaning. This should usually be called when a
 * file is updated or deleted.
 *
 * Need to lock the band before calling append_deletelog because meminfo
 * is changed in append_deletelog
 *
 * @param inode Inode of the file that is being deleted/updated.
 */
inline void
mark_for_cleaning(inode_t inode) {
    int fd_log;


    /* Ensure that the file was actually written to the Shingled partition */
    if (inode.entrynumber <= 0)
        return;

    dod((fd_log = init_deletelog(inode.band_id)) > 0);
    append_deletelog(inode.band_id, fd_log, inode.band_offset,
                     inode.size, inode.entrynumber);
    close(fd_log);
    log(LOG_SFS_INFO, "Delete Log entry: Band %ld, RBA %ld, %d blocks, "
        "Append Log Entry: %d!\n", inode.band_id,
        inode.band_offset, inode.size, inode.entrynumber);
}



/**
 * @brief Frees the specified inode.
 *
 * This function should be called when an inode needs to be freed. This usually
 * occurs when a file has been deleted or renamed (when a file is renamed to an
 * existing file).
 *
 * @param inodeno The inode no. of the inode to be freed
 */
inline void
inode_free(int inodeno) {
    //inode_t inode;

    /* Zero out the inode, except for the inode no. */
    //memset(&inode, 0, sizeof(inode_t));
    //inode.inode_num = inodeno;

    /* Write out the inode data */
    //if (put_inode(inodeno, &inode))
    //    exit(EXIT_FAILURE);

    /* Remove the xattr_inode file */
    /*char inode_xattr_path[PATH_MAX];
    xattr_path_from_inode(SFSDATA.path_unshingled, inode_xattr_path, inodeno);
    assert(!unlink(inode_xattr_path));*/

    /* Mark the inode as free in the inode bitmap */
    put_freeinodenum(inodeno);
}

inline void
inode_xattr_free(int inodeno) {
    char inode_xattr_path[PATH_MAX];
    xattr_path_from_inode(SFSDATA.path_unshingled, inode_xattr_path, inodeno);
    assert(!unlink(inode_xattr_path));
}

/**
 * @brief Updates the band usage data in the inode
 *
 * @param inode Inode of the file that is being updated
 * @param band The updated band no.
 * @param rba The updated RBA
 * @param size The new size/extent of the file
 */
inline void
inode_update(inode_t *inode, band_t band, rba_t rba, size_t size) {
    int fd_log;


    /* Update the band no. and offset in the file's inode */
    inode->band_id = band;
    inode->band_offset = rba;
    inode->size = size;

    /* Update the append log for the band */
    dod((fd_log = init_storelog(band)) > 0);
    inode->entrynumber = append_storelog(band, fd_log, rba, size, inode->inode_num, 0);
    close(fd_log);

    /* Write the updated inode to disk */
    dod(!put_inode(inode->inode_num, inode));
    log(LOG_SFS_INFO, "Append log entry for band %d at %d\n",
        band, inode->entrynumber);
}



/**
 * @brief Unlinks the inode
 *
 * This function marks the space described by the inode for cleaning and then
 * frees the inode.
 *
 * add inode list lock. Check after lock the inode structure is the same.
 * If not, release lock and return failure
 *
 * @param inodeno Inode number of the file that is being deleted
 * @param inode Inode structure that is return by get_inode before getting
 *        the inodelist lock
 *
 * @return 0 on success
 *         -1 on failure
 */
inline int
inode_unlink(int inodeno, inode_t inode) {
    inode_t tempinode;
    band_meminfo_lock_band_for_write( inode.band_id);
    /* lock inode table*/
    lock_inodelist();
    get_inode(inodeno, &tempinode);
    if (tempinode.band_id == inode.band_id)
    {
        /* Mark the space described by the inode for cleaning */
        mark_for_cleaning(inode);

        /* Free the inode */
        inode_free(inode.inode_num);
        /* release inode table lock*/
        unlock_inodelist();
        band_meminfo_release_band_lock(inode.band_id);
        return 0;
    }
    else
    {
        /* release inode table lock */
        unlock_inodelist();
        /* release band lock */
        band_meminfo_release_band_lock(inode.band_id);
        return -1;
    }
}
