/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file inode.c
 *
 * @brief Implements the inode abstraction for ShingledFS.
 *
 * @author Jainam Shah <jashah@andrew.cmu.edu>
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include "helper.h"
#include "common.h"
#include <errno.h>

#include "inode.h"


#define FILE_INODE_BITMAP               "/.smr_inode_bitmap"
#define FILE_INODE_LIST                 "/.smr_inode_list"

static int fd_bitmap;
static int fd_list;

pthread_rwlock_t inodelock;

/*
 * @brief Fetchs the inode information from the inode file
 *
 * @param inode_num The INode number
 * @param node Pointer to buffer to hold the inode data
 *
 * @return 0 on success; -1 on error
 */

int get_inode (int inode_num, inode_t *node)
{
    char inode_xattr_path[PATH_MAX];
    xattr_path_from_inode(SFSDATA.path_unshingled, inode_xattr_path, inode_num);
    if(getxattr(inode_xattr_path, X_INODE_STRUCT, node, sizeof(inode_t)) <= 0) {
	dbg("Error reading inode in get_inode \n");
	return -1;
    }
    return 0;
}



/*
 * @brief Writes inode data to the inodes file
 *
 * @param inode_num The INode number
 * @param node Pointer to buffer holding the inode data
 *
 * @return 0 on success; -1 on error
 */

int put_inode (int inode_num, inode_t *node)
{
    char inode_xattr_path[255];
    xattr_path_from_inode(SFSDATA.path_unshingled, inode_xattr_path, inode_num);
    assert(!setxattr(inode_xattr_path, X_INODE_STRUCT, node, sizeof(inode_t), 0));
    return 0;
}



/**
 * @brief Returns the offset of an inode record in the inodes file
 *
 * @param inode_num The INode number
 *
 * @return Offset in inode  in the inodes file
 */

int get_offset(int inode_num)
{
    return (sizeof(inode_t) * (inode_num - 1));
}



/**
 * @brief Initializes the inode bitmap
 *
 * @return File descriptor of the Inode Bitmap file on success.
 *          -1 on failure
 */

int init_inodebitmap(void)
{
    char path[PATH_MAX];
    int max_inodes;
    char bit = 0;
    int i;


    translate_path(FILE_INODE_BITMAP, SFSDATA.path_unshingled, path);

    fd_bitmap = open(path, O_RDWR | O_CREAT | O_EXCL, 0644);

    if (fd_bitmap < 0) {
        //inode bitmap already exists.
        fd_bitmap = open(path, O_RDWR);
        if (fd_bitmap < 0) {
            dbg("Error opening inode bitmap file \n");
            return -1;
        }
        return 0;
    }


    max_inodes = MAX_NUM_INODES / 8;

    for (i = 0; i < max_inodes; i++) {
        if (pwrite(fd_bitmap, &bit, 1, i) < 0) {
            dbg("Error writing to inode bitmap in init_inodebitmap \n");
            return -1;
        }
    }

    return 0;
}


/**
 * @brief Initializes the inode list.
 *
 * If the inode list already exists then will open the inode list file and
 * return the file descriptor.
 *
 * @param : None.
 *
 * @return fd : File descriptor on Success.
 *         -1 : on Error.
 *
 */

int init_inodelist(void)
{
    char path[PATH_MAX];
    int max_inodes;
    int i;
    inode_t node;


    translate_path(FILE_INODE_LIST, SFSDATA.path_unshingled, path);

    fd_list = open(path, O_RDWR | O_CREAT | O_EXCL, 0644);

    if (fd_list < 0) {
        //inode list already exists.
        fd_list = open(path, O_RDWR);
        if (fd_list < 0) {
            dbg("Error opening inode list file \n");
            return -1;
        }
        return 0;
    }
    else {
        if(pthread_rwlock_init(&inodelock, NULL) != 0)
            return -1;
    }

    max_inodes = MAX_NUM_INODES / 8;
    node.avail = 0;

    for (i = 0; i < max_inodes; i++) {
        node.inode_num = i + 1;
        if (pwrite(fd_list, &node, sizeof(inode_t),
                   i * sizeof(inode_t)) < 0) {
            return -1;
        }
    }

    return 0;
}


/*
 * @brief : Get next free inode and update the inode bitmap.
 *
 * @param : None.
 *
 * @return inode number : on Success.
 *         0 : if no inodes are free.
 *
 */

int get_freeinodenum(void)
{
    return ++max_inode_used;
    /*int i,j;
    int max_inodes;
    int inode_num = 0;
    char bit;

    max_inodes = MAX_NUM_INODES / 8;

    for (i = 0; i < max_inodes; i++) {
        if (pread(fd_bitmap, &bit, 1, i) < 0) {
            return 0;
        }

        if(bit == 0xFF) {
            continue;
        }

        for (j = 0; j < 8; j++) {

            if ((bit & (1 << j)) == 0) {
                bit = bit | (1 << j);
                if (pwrite(fd_bitmap, &bit, 1, i) < 0) {
                    dbg("Error writing to inode bitmap in get_freeinodenum\n");
                    exit(1);
                }

                inode_num = i * 8 + j + 1;
                break;
            }
        }
        if (inode_num != 0) {
            break;
        }
    }

    return inode_num;
    */
}


/*
 * @brief : Free the inode by updating the inode bitmap.
 *
 * @param inode_num : inode number.
 *
 * @return : None.
 *
 */

void put_freeinodenum(int inode_num)
{
    /*int i, j;
    char bit;

    if (inode_num <= 0) {
        return;
    }

    i = (inode_num - 1) / 8;
    j = (inode_num - 1) % 8;


    if (pread(fd_bitmap, &bit, 1, i) < 0) {
        dbg("Error reading inode bitmap in put_freeinodenum\n");
        exit(1);
    }
    bit = bit & (~(1 << j));
    if (pwrite(fd_bitmap, &bit, 1, i) < 0) {
        dbg("Error writing inode bitmap in put_freeinodenum \n");
        exit(1);
    }
    */
}

void lock_inodelist(void)
{
    dbg("request lock inodelist\n");
    assert(pthread_rwlock_wrlock(&inodelock) == 0);
    dbg("success: lock inodelist\n");
}

void unlock_inodelist(void)
{
    dbg("request unlock inodelist\n");
    assert(pthread_rwlock_unlock(&inodelock) == 0);
    dbg("success: unlock inodelist\n");
}
