/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file shingledfs.c
 *
 * @brief The Shingled File System
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#define FUSE_USE_VERSION 26


#include <fuse.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/xattr.h>
#include <limits.h>

#include "helper.h"
#include "shingledfs.h"



/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

#ifdef _EDI_ON_RAW
#define ARG_COUNT_MIN                   5
#else
#define ARG_COUNT_MIN                   4
#endif

#define ARG_PATH_UNSHINGLED             1
#define ARG_PATH_TMPFS                  2
#define ARG_PATH_EDI                    3
#ifdef _EDI_ON_RAW
#define ARG_PATH_EDI_METADATA           4
#endif

#define ARG_SINGLE_THREADED             "-s"
#define ARG_FOREGROUND_EXEC             "-f"
//#define ARG_FSNAME                            "-ofsname=ShingledFS"



/* ====================== */
/* -- Static Variables -- */
/* ====================== */

/**
 * Functions supported by ShingledFS
 */
struct fuse_operations shingledfs_ops = {
    .init        = shingledfs_init,
    .destroy     = shingledfs_destroy,

    .mkdir       = shingledfs_mkdir,
    .rmdir       = shingledfs_rmdir,
    .opendir     = shingledfs_opendir,
    .readdir     = shingledfs_readdir,
    .releasedir  = shingledfs_releasedir,

    .mknod       = shingledfs_mknod,
    .open        = shingledfs_open,
    .read        = shingledfs_read,
    .write       = shingledfs_write,
    .flush       = shingledfs_flush,
    .release     = shingledfs_release,
    .rename      = shingledfs_rename,
    .unlink      = shingledfs_unlink,
    .truncate    = shingledfs_truncate,

    .getattr     = shingledfs_getattr,
    .fgetattr    = shingledfs_fgetattr,
    .setxattr    = shingledfs_setxattr,
    .getxattr    = shingledfs_getxattr,
    .listxattr   = shingledfs_listxattr,
    .removexattr = shingledfs_removexattr,

    .access      = shingledfs_access,
    .utime       = shingledfs_utime,
    .chmod       = shingledfs_chmod,
    .chown       = shingledfs_chown,

    .symlink     = shingledfs_symlink,
    .readlink    = shingledfs_readlink,
    .link        = shingledfs_link,
    .statfs      = shingledfs_statfs,

//    .ftruncate   = shingledfs_ftruncate,
    .fsync       = shingledfs_fsync,
//    .fsyncdir    = NULL,
};



/* ================================================= */
/* -- Usage and Command-line Arguments Processing -- */
/* ================================================= */

/**
 * @brief Prints the usage format for launching/mounting ShingledFS.
 */
static void
usage(const char *program_name)
{
    fprintf(stderr,
            "USAGE: %s <Unshingled path> <tmpFS path> <EDI path> <FUSE mount path>\n",
            program_name);
    abort();
}



/**
 * @brief Processes the incoming arguments.
 *
 * This function stores the required paths, namely the path to the unshingled
 * partition and to the EDI file in the ShingledFS Data structure for later use.
 *
 * It then removes the extra arguments, adjusts the argument count and makes the
 * necessary changes to the argument list to pass it to FUSE.
 *
 * If this function finds an error with the arguments, then it terminates the
 * application.
 *
 * @param argc Pointer to the count of arguments
 * @param argv Array of command line arguments
 */
static void
process_args(int *argc,
             char *argv[]) {
    int i, arg_error = 0;
    int is_singled_threaded = 0;


    /* Ensure that the arguments are available */
    if (*argc < ARG_COUNT_MIN)
        usage(argv[0]);


    /* Strip out any terminating /'s in the path to the directories */
    i = strlen(argv[ARG_PATH_UNSHINGLED]) - 1;
    if (argv[ARG_PATH_UNSHINGLED][i] == '/')
        argv[ARG_PATH_UNSHINGLED][i] = '\0';

    i = strlen(argv[ARG_PATH_TMPFS]) - 1;
    if (argv[ARG_PATH_TMPFS][i] == '/')
        argv[ARG_PATH_TMPFS][i] = '\0';


    /* Validate the path to the Unshingled partition */
    if (!is_valid_dir(argv[ARG_PATH_UNSHINGLED])) {
        arg_error = 1;
        fprintf(stderr,
                "Error accessing %s. Please ensure it is a valid directory!\n",
                argv[ARG_PATH_UNSHINGLED]);
    }

    /* Validate the path to the tmpFS partition */
    if (!is_valid_dir(argv[ARG_PATH_TMPFS])) {
        arg_error = 1;
        fprintf(stderr,
                "Error accessing %s. Please ensure it is a valid directory!\n",
                argv[ARG_PATH_TMPFS]);
    }

#ifndef _EDI_ON_RAW
    /* Validate the path to the Emulated Disk Image */
    if (!is_valid_file(argv[ARG_PATH_EDI])) {
        arg_error = 1;
        fprintf(stderr,
                "Error accessing %s. Please ensure it is a valid file!\n",
                argv[ARG_PATH_EDI]);
    }
#else
    /* Validate the path to the Emulated Disk Image Metadata file */
    if (!is_valid_file(argv[ARG_PATH_EDI_METADATA])) {
	arg_error = 1;
	fprintf(stderr,
                "Error accessing %s. Please ensure it is a valid file!\n",
                argv[ARG_PATH_EDI_METADATA]);
    }
#endif

    /* Check if there are any errors */
    if (arg_error)
        usage(argv[0]);


    /* Store the arguments */
    strcpy(shingledfs_data.path_unshingled, argv[ARG_PATH_UNSHINGLED]);
    strcpy(shingledfs_data.path_tmpfs, argv[ARG_PATH_TMPFS]);
    strcpy(shingledfs_data.path_edi, argv[ARG_PATH_EDI]);
#ifdef _EDI_ON_RAW
    strcpy(shingledfs_data.path_edi_metadata, argv[ARG_PATH_EDI_METADATA]);
#endif

    dbg("Unshingled path: %s\n", shingledfs_data.path_unshingled);
    dbg("tmpFS path: %s\n", shingledfs_data.path_tmpfs);
    dbg("EDI path: %s\n", shingledfs_data.path_edi);
#ifdef _EDI_ON_RAW
    dbg("EDI metadata path: %s\n", shingledfs_data.path_edi_metadata);
#endif

    /* Adjust the argument count and remove the Shingled FS arguments */
    *argc -= ARG_COUNT_MIN - 1;
    for(i=0; i<(*argc); i++) {
        /* Move the argument forward */
        argv[i] = argv[i + ARG_COUNT_MIN - 1];

        /* Check if the argument is the -s flag for single-threaded execution */
        if (strcasecmp(argv[i], ARG_SINGLE_THREADED) == 0)
            is_singled_threaded = 1;
    }

    /* @todo Ensure that the File System Name is passed */
//      argv[(*argc)++] = ARG_FSNAME;

    /* Ensure that the process is running in single-threaded mode */
    if (!is_singled_threaded)
        argv[(*argc)++] = ARG_SINGLE_THREADED;

#ifdef DEBUG
    /* When running in debug mode, ensure that FUSE runs in the foreground */
    argv[(*argc)++] = ARG_FOREGROUND_EXEC;
#endif


    /* Null-terminate the arguments array */
    argv[*argc] = NULL;
}



/* =================== */
/* -- Main Function -- */
/* =================== */

int
main(int argc,
     char *argv[]) {
    /* Process the arguments */
    process_args(&argc, argv);

    /* Reset the umask to ensure all permissions get set exactly as specified */
    umask(0);

    /* Kick start FUSE */
    return fuse_main(argc, argv, &shingledfs_ops, &shingledfs_data);
}
