/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file shingledfs.h
 *
 * @brief Macros, data structures and function prototypes pertaining to the
 * Shingled File System.
 *
 * @author Anand Suresh <anandsuresh@cmu.edu>
 */

#ifndef __SHINGLEDFS_H_
#define __SHINGLEDFS_H_


#include <fuse.h>
#ifdef __cplusplus
 extern "C" {
#endif
#include "common.h"


/* ======================= */
/* -- Macro Definitions -- */
/* ======================= */

/**
 * @brief Maximum length for the name of a file
 */
#define MAX_FILENAME_LEN        255



/* ========================= */
/* -- Function Prototypes -- */
/* ========================= */

shingledfs_data_t *shingledfs_init(struct fuse_conn_info *conn);
void shingledfs_destroy(void *userdata);

int shingledfs_mkdir(const char *path, mode_t mode);
int shingledfs_rmdir(const char *path);
int shingledfs_opendir(const char *path, struct fuse_file_info *fi);
int shingledfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                       off_t offset, struct fuse_file_info *fi);
int shingledfs_releasedir(const char *path, struct fuse_file_info *fi);


int shingledfs_mknod(const char *path, mode_t mode, dev_t dev);
int shingledfs_open(const char *path, struct fuse_file_info *fi);
int shingledfs_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi);
int shingledfs_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi);
int shingledfs_flush(const char *path, struct fuse_file_info *fi);
int shingledfs_release(const char *path, struct fuse_file_info *fi);
int shingledfs_rename(const char *path, const char *path_new);
int shingledfs_unlink(const char *path);
int shingledfs_truncate(const char *path, off_t offset);

int shingledfs_getattr(const char *path, struct stat *statbuf);
int shingledfs_fgetattr(const char *path, struct stat *statbuf,
                        struct fuse_file_info *fi);
int shingledfs_setxattr(const char *path, const char *name, const char *value,
                        size_t size, int flags);
int shingledfs_getxattr(const char *path, const char *name, char *value,
                        size_t size);
int shingledfs_listxattr(const char *path, char *list, size_t size);
int shingledfs_removexattr(const char *path, const char *name);

int shingledfs_access(const char *path, int mask);
int shingledfs_utime(const char *path, struct utimbuf *tv);
int shingledfs_chmod(const char *path, mode_t mode);
int shingledfs_chown(const char *path, uid_t uid, gid_t gid);

int shingledfs_symlink(const char *path, const char *path_target);
int shingledfs_readlink(const char *path, char *buf, size_t size);
int shingledfs_link(const char *path, const char *path_target);

int shingledfs_statfs(const char *path, struct statvfs *statv);
int shingledfs_fsync(const char * path, int datasync, struct fuse_file_info *fi);


#ifdef __cplusplus
}
#endif
#endif /* __SHINGLEDFS_H_ */
