/**
 * Copyright 2013 Parallel Data Lab (Carnegie Mellon University)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file    stats.c
 * A module to collect/store the current state of ShingledFS.
 *
 * Responsibilities include
 *  - measuring the idle
 *  - invoking the cleaner to clean bands
 *  - keeping track of all the filesystem stats
 *  - serving the stats to users
 *
 * @author  Pavan Kumar ALampalli (pavankumar@cmu.edu)
 * @date    12-oct-2012
 */

#include "stats.h"
#include "assert.h"
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "band_meminfo.h"

/** Dumps all stats to this file every stats_update() */
#define STATS_FILE_PATH "/tmp/stats.csv"
static FILE* statsfp;
static FILE* newstatsfp;
static FILE* cleanerfp;
static char * stats_desc[] = {
    "Last disk activity",
    "Free Blocks",
    "Max. Blocks",
    "Cleaned Blocks",
    "Cumulated Cleaned Blocks",
    "Moved Blocks",
    "Cumulated Moved Blocks"};

/** Datastructure to store all the counters used by the fs */
unsigned long int stats[MAX_COUNTERS];

inline void stats_set(Counter c, unsigned long value)
{
    stats[c] = value;
}

inline void stats_incr(Counter c, unsigned long value)
{
    stats[c] += value;
}

inline void stats_decr(Counter c, unsigned long value)
{
    assert(stats[c] >= value);
    stats[c] -= value;
}

unsigned long stats_get(Counter c)
{
    return stats[c];
}


int stats_update(void)
{
    char filename[50];
    int i, len;
    time_t curtime;
    
    if (!statsfp) {
        curtime = time(NULL);

        /* Initialize the stats file */
        len = strftime(filename, 50, "/tmp/smr%Y%m%d%H%M%S_stats.csv",
                       localtime(&curtime));
        if (!len) {
            fprintf(stderr, "Error generating log file name: %s\n", filename);
            return -1;
        }
        
        statsfp = fopen(filename, "w");
        if (!statsfp) {
            fprintf(stderr, "Error creating file: %s\n", filename);
            return -1;
        }
        /* print headers */
        for (i = 0;i < MAX_COUNTERS;i++) {
            fprintf(statsfp, "%s\t", stats_desc[i]);
        }
        fprintf(statsfp,"\n");
    }

    for (i = 0;i < MAX_COUNTERS;i++) {
        fprintf(statsfp, "%zu\t", stats[i]);
    }
    fprintf(statsfp,"\n");

    fflush(statsfp);
    return 0;
}                   

int stats_disk_utilization(void)
{
    int total_blocks = stats_get(DISK_MAX_BLKS);
    int used_blocks = total_blocks - stats_get(DISK_FREE_BLKS);
    int fs_full = (used_blocks*100.0) / total_blocks;
    return fs_full;
}

void output_to_stats( band_data_t * band_i, Clean_time clean_time, int clean_flag, char * c, int * length)
{
    if(clean_time == BEFORE)
    {
        if(c == NULL)
            return;
        if(band_i->state_new == 1)
            *length += sprintf(c+(*length), "%d\t%zu\t", 0, band_i->used);
        else
            *length += sprintf(c+(*length), "%zu\t%d\t", band_i->used, 0);
        if(band_i->state_delete == 1)
            *length += sprintf(c+(*length), "%d\t%zu\t", 0, band_i->deleted);
        else
            *length += sprintf(c+(*length), "%zu\t%d\t", band_i->deleted, 0);
            *length += sprintf(c+(*length), "%zu\t%d\t", band_i->free, 0);
    }
    else if(clean_time == AFTER && clean_flag == 1)
    {
        if(band_i->state_clean_to == 1)
            fprintf(newstatsfp, "%d\t%zu\t", 0, band_i->used);
        else
            fprintf(newstatsfp, "%zu\t%d\t", band_i->used, 0);
        if(band_i->state_delete == 1)
            fprintf(newstatsfp, "%d\t%zu\t", 0, band_i->deleted);
        else
            fprintf(newstatsfp, "%zu\t%d\t", band_i->deleted, 0);
        if(band_i->state_cleaned == 1)
            fprintf(newstatsfp, "%d\t%zu\t", 0, band_i->free);
        else
            fprintf(newstatsfp, "%zu\t%d\t", band_i->free, 0);
        band_meminfo_reset_state();
    }
    else if(clean_time == AFTER && clean_flag ==0)
    {
    //no cleaning for this time. Don't need to output to stats to file
    }
    else 
    {
        fprintf(stderr, "cleaning time error, not before or after");
        fprintf(newstatsfp, "error");
    }
}
        
void stats_cleaner(Clean_time clean_time, int clean_flag, char * c)
{
    band_data_t * band_i;
    band_data_t * band;
    char cleanerfn[100] = "\0";
    char cleaner[] = "/cleanerstats.txt";
    char newstatsfn[100] = "\0";
    char newstats[] = "/newstats.txt";
    time_t c_time = 0;
    int length = 0;
    char *statspath = NULL;
    if((statspath = getenv("STATSPATH")) == NULL)
        fprintf(stderr, "Null returned from getenv STATSPATH");

    strcat(cleanerfn, statspath);
    strcat(cleanerfn, cleaner);
    strcat(newstatsfn, statspath);
    strcat(newstatsfn, newstats);

    if (!cleanerfp) {
        cleanerfp = fopen(cleanerfn, "w");
        if (!cleanerfp) {
            fprintf(stderr, "Error creating file: %s\n", cleanerfn);
            return;
        }
    }
    if (!newstatsfp) {
        newstatsfp = fopen(newstatsfn, "w");
        if (!newstatsfp) {
            fprintf(stderr, "Error creating file: %s\n", newstatsfn);
            return;
        }
        //fprintf(newstatsfp, "%s\t%s\n","No.", "amount cleaned"); 
    }
    rewind(cleanerfp);
    fprintf(cleanerfp, "%s\t%s\t%s\t%s\n","No.", "u_band", "deleted", "time"); 
    c_time = time(NULL);
    if((clean_time == BEFORE) && (c!=NULL))
    	length += sprintf(c + length, "BEFORE%lu\t", c_time);
    else if((clean_time == AFTER) && (clean_flag == 1) && (c != NULL)) {
        fprintf(newstatsfp, "%s", c);
    	fprintf(newstatsfp, "AFTER%lu\t", c_time);
    }
    else if((clean_time == AFTER) && (clean_flag == 0))
    {
    //no cleaning happened this time. Do not output to file
    }
    else
    	fprintf(stderr, "UNKNOWN%lu %d\t", c_time, clean_flag);
    band =(band_data_t *) malloc(SFSDATA.edi_bandcount * sizeof(band_data_t));
    if (band == NULL) {
        free(band);
        fprintf(stderr,"fail to malloc for band_i\n");
    	fprintf(newstatsfp, "fail to malloc for band_i\n");
        return;
    }
    int uband = 0;//utilization of a band (used/band_size)
    int deleted_per = 0;//fraction of band that is available to cleaner
                        //(deleted/band_size)
    int i = 0;

    if (band_meminfo_get_cleaner_stats(band) != 0){
        fprintf(stderr, "error get stats from band_meminfo for cleaner stats\n");
    	fprintf(newstatsfp, "error get stats from band_meminfo\n");
        free(band);
        return;
    }
    for (i = 1; i < SFSDATA.edi_bandcount; i++) {
        band_i = &(band[i]);
        uband = band_i->used*100/band_i->band_size;
        deleted_per = band_i->deleted*100/band_i->band_size;
        fprintf(cleanerfp, "%d\t %d\t %d\t %s\n", i, uband, deleted_per, ctime(&(band_i->last_mod_time)));
        assert(band_i->used + band_i->deleted + band_i->free == band_i->band_size);
        output_to_stats(band_i, clean_time, clean_flag, c, &length);
    }
    free(band);
    //print out amount cleaned and amount moved in this run at the end of the line
    if(clean_time == BEFORE)
    length += sprintf(c+length, "SUM %lu \t %lu \t %lu \t %lu\n",
	 stats[MOVED_BLKS], stats[CLEANED_BLKS], stats[CUM_MOVED_BLKS], 
	 stats[CUM_CLEANED_BLKS]);
    else if( clean_time == AFTER && clean_flag == 1)
    fprintf(newstatsfp, "SUM %lu \t %lu \t %lu \t %lu\n",
	 stats[MOVED_BLKS], stats[CLEANED_BLKS], stats[CUM_MOVED_BLKS], 
	 stats[CUM_CLEANED_BLKS]);
    fflush(cleanerfp);
    fflush(newstatsfp);
    return;
}
